import collections
import itertools
from typing import Dict, List, Optional, Set, Tuple

from relik.inference.data.splitters.blank_sentence_splitter import BlankSentenceSplitter
from relik.inference.data.splitters.base_sentence_splitter import BaseSentenceSplitter
from relik.inference.data.splitters.window_based_splitter import WindowSentenceSplitter
from relik.inference.data.splitters.spacy_sentence_splitter import SpacySentenceSplitter
from relik.inference.data.tokenizers.base_tokenizer import BaseTokenizer
from relik.reader.data.relik_reader_sample import RelikReaderSample
from relik.inference.data.objects import AnnotationType, TaskType


class WindowManager:
    def __init__(
        self, tokenizer: BaseTokenizer, splitter: BaseSentenceSplitter | None = None
    ) -> None:
        self.tokenizer = tokenizer
        self.splitter = splitter or BlankSentenceSplitter()

    def create_windows(
        self,
        documents: str | List[str],
        window_size: int | None = None,
        stride: int | None = None,
        max_length: int | None = None,
        doc_ids: str | int | List[str] | List[int] | None = None,
        doc_topic: str | None = None,
        is_split_into_words: bool = False,
        mentions: List[List[List[int]]] = None,
        annotation_type: str | AnnotationType = AnnotationType.CHAR,
        task_type: str | TaskType = TaskType.SPAN,
    ) -> Tuple[List[RelikReaderSample], List[RelikReaderSample]]:
        """
        Create windows from a list of documents.

        Args:
            documents (:obj:`str` or :obj:`List[str]`):
                The document(s) to split in windows.
            window_size (:obj:`int`):
                The size of the window.
            stride (:obj:`int`):
                The stride between two windows.
            max_length (:obj:`int`, `optional`):
                The maximum length of a window.
            doc_id (:obj:`str` or :obj:`int`, `optional`):
                The id of the document(s).
            doc_topic (:obj:`str`, `optional`):
                The topic of the document(s).
            is_split_into_words (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether the input is already pre-tokenized (e.g., split into words). If :obj:`False`, the
                input will first be tokenized using the tokenizer, then the tokens will be split into words.
            mentions (:obj:`List[List[List[int]]]`, `optional`):
                The mentions of the document(s).

        Returns:
            :obj:`List[RelikReaderSample]`: The windows created from the documents.
        """
        # normalize input
        if isinstance(documents, str):
            documents = [documents]

        if doc_ids is not None:
            if isinstance(doc_ids, int):
                doc_ids = [doc_ids]

            # check doc_id and documents have the same length
            if len(doc_ids) != len(documents):
                raise ValueError(
                    f"doc_id and documents must have the same length. Found {len(doc_ids)} doc_id and {len(documents)} documents."
                )

        # batch tokenize
        documents_tokens = self.tokenizer(
            documents, is_split_into_words=is_split_into_words
        )

        if (
            task_type == TaskType.SPAN
            and annotation_type == AnnotationType.WORD
            and mentions is not None
        ):
            # we need to move the mentions to the char level
            mentions = [
                [
                    [
                        documents_tokens[i][m[0]].idx,
                        documents_tokens[i][m[1] - 1].idx
                        + len(documents_tokens[i][m[1] - 1].text),
                    ]
                    for m in doc_mentions
                ]
                for i, doc_mentions in enumerate(mentions)
            ]
        elif (
            task_type == TaskType.TRIPLET
            and annotation_type == AnnotationType.CHAR
            and mentions is not None
        ):
            # we need to move the mentions to the word level, i.e. find the start of the word with character m[0] and the end of the word with character m[1]
            mentions = [
                [
                    [
                        next(
                            i for i, token in enumerate(doc_tokens) if token.idx >= m[0]
                        ),
                        next(
                            i
                            for i, token in enumerate(doc_tokens)
                            if token.idx + len(token.text) >= m[1]
                        ),
                    ]
                    for m in doc_mentions
                ]
                for doc_tokens, doc_mentions in zip(documents_tokens, mentions)
            ]

        # set splitter params
        if hasattr(self.splitter, "window_size"):
            self.splitter.window_size = window_size or self.splitter.window_size
        if hasattr(self.splitter, "window_stride"):
            self.splitter.window_stride = stride or self.splitter.window_stride

        windowed_documents, windowed_blank_documents = [], []

        # TODO: normalize mantions and return type, check when doc_ids is None
        if mentions is not None:
            assert len(documents) == len(
                mentions
            ), f"documents and mentions should have the same length, got {len(documents)} and {len(mentions)}"
            doc_iter = zip(
                documents, documents_tokens, doc_ids or itertools.repeat([]), mentions
            )
        else:
            doc_iter = zip(
                documents,
                documents_tokens,
                doc_ids or itertools.repeat([]),
                itertools.repeat([]),
            )

        for infered_doc_id, (
            document,
            document_tokens,
            doc_id,
            document_mentions,
        ) in enumerate(doc_iter):
            if doc_topic is None:
                doc_topic = document_tokens[0].text if len(document_tokens) > 0 else ""

            if not doc_id:
                doc_id = infered_doc_id

            # if doc_id is None:
            # actual_doc_id = doc_ids or infered_doc_id

            splitted_document = self.splitter(document_tokens, max_length=max_length)

            document_windows = []
            for window_id, window in enumerate(splitted_document):
                window_text_start = window[0].idx
                window_text_end = window[-1].idx + len(window[-1].text)
                if isinstance(document, str):
                    text = document[window_text_start:window_text_end]
                else:
                    # window_text_start = window[0].idx
                    # window_text_end = window[-1].i
                    text = " ".join([w.text for w in window])

                if annotation_type == AnnotationType.WORD:
                    spans = [
                        [m[0] - window[0].i, m[1] - window[0].i]
                        for m in document_mentions
                        if window[-1].i > m[0] >= window[0].i
                        and window[-1].i >= m[1] >= window[0].i
                    ]
                else:
                    spans = [
                        [m[0], m[1]]
                        for m in document_mentions
                        if window_text_end > m[0] >= window_text_start
                        and window_text_end >= m[1] >= window_text_start
                    ]
                sample = RelikReaderSample(
                    doc_id=doc_id,
                    window_id=window_id,
                    text=text,
                    tokens=[w.text for w in window],
                    words=[w.text for w in window],
                    doc_topic=doc_topic,
                    offset=window_text_start,
                    spans=spans,
                    token2char_start={str(i): w.idx for i, w in enumerate(window)},
                    token2char_end={
                        str(i): w.idx + len(w.text) for i, w in enumerate(window)
                    },
                    char2token_start={str(w.idx): w.i for i, w in enumerate(window)},
                    char2token_end={
                        str(w.idx + len(w.text)): w.i for i, w in enumerate(window)
                    },
                )
                if mentions is not None and len(sample.spans) == 0:
                    windowed_blank_documents.append(sample)
                else:
                    document_windows.append(sample)

            windowed_documents.extend(document_windows)
        if mentions is not None:
            return windowed_documents, windowed_blank_documents, documents_tokens
        else:
            return windowed_documents, [], documents_tokens

    def merge_windows(
        self, windows: List[RelikReaderSample]
    ) -> List[RelikReaderSample]:
        windows_by_doc_id = collections.defaultdict(list)
        for window in windows:
            windows_by_doc_id[window.doc_id].append(window)

        merged_window_by_doc = {
            doc_id: self._merge_doc_windows(doc_windows)
            for doc_id, doc_windows in windows_by_doc_id.items()
        }

        return list(merged_window_by_doc.values())

    def _merge_doc_windows(self, windows: List[RelikReaderSample]) -> RelikReaderSample:
        if len(windows) == 1:
            return self._normalize_single_window(windows[0])

        # if not isinstance(self.splitter, WindowSentenceSplitter):
        #     # here we don't really need to merge windows, just normalize them
        #     # TODO: check if we need to merge windows in this case
        #     return [self._normalize_single_window(w) for w in windows]

        if len(windows) > 0 and getattr(windows[0], "offset", None) is not None:
            windows = sorted(windows, key=(lambda x: x.offset))

        window_accumulator = windows[0]

        for next_window in windows[1:]:
            window_accumulator = self._merge_window_pair(
                window_accumulator, next_window
            )

        return window_accumulator

    def _normalize_single_window(self, window: RelikReaderSample) -> RelikReaderSample:
        # normalize the `span_candidates` and `triplet_candidates` fields
        # to mimic the behavior of the `merge_windows` method

        if (
            getattr(window, "span_candidates", None) is not None
            and len(window.span_candidates) > 0
        ):
            window._d["span_candidates"] = [window.span_candidates]

        if (
            getattr(window, "triplet_candidates", None) is not None
            and len(window.triplet_candidates) > 0
        ):
            window._d["triplet_candidates"] = [window.triplet_candidates]

        return window

    @staticmethod
    def _merge_tokens(
        window1: RelikReaderSample, window2: RelikReaderSample
    ) -> Tuple[list, dict, dict]:
        w1_tokens = window1.tokens[1:-1]
        w2_tokens = window2.tokens[1:-1]

        # find intersection if any
        tokens_intersection = 0
        for k in reversed(range(1, len(w1_tokens))):
            if w1_tokens[-k:] == w2_tokens[:k]:
                tokens_intersection = k
                break

        final_tokens = (
            [window1.tokens[0]]  # CLS
            + w1_tokens
            + w2_tokens[tokens_intersection:]
            + [window1.tokens[-1]]  # SEP
        )

        w2_starting_offset = len(w1_tokens) - tokens_intersection

        def merge_char_mapping(t2c1: dict, t2c2: dict) -> dict:
            final_t2c = dict()
            final_t2c.update(t2c1)
            for t, c in t2c2.items():
                t = int(t)
                if t < tokens_intersection:
                    continue
                final_t2c[str(t + w2_starting_offset)] = c
            return final_t2c

        return (
            final_tokens,
            merge_char_mapping(window1.token2char_start, window2.token2char_start),
            merge_char_mapping(window1.token2char_end, window2.token2char_end),
        )

    @staticmethod
    def _merge_words(
        window1: RelikReaderSample, window2: RelikReaderSample
    ) -> Tuple[list, dict, dict]:
        w1_words = window1.words
        w2_words = window2.words

        # find intersection if any
        words_intersection = 0
        for k in reversed(range(1, len(w1_words))):
            if w1_words[-k:] == w2_words[:k]:
                words_intersection = k
                break

        final_words = w1_words + w2_words[words_intersection:]

        w2_starting_offset = len(w1_words) - words_intersection

        def merge_word_mapping(t2c1: dict, t2c2: dict) -> dict:
            final_t2c = dict()
            if t2c1 is None:
                t2c1 = dict()
            if t2c2 is None:
                t2c2 = dict()
            final_t2c.update(t2c1)
            for t, c in t2c2.items():
                t = int(t)
                if t < words_intersection:
                    continue
                final_t2c[str(t + w2_starting_offset)] = c
            return final_t2c

        return (
            final_words,
            merge_word_mapping(window1.token2word_start, window2.token2word_start),
            merge_word_mapping(window1.token2word_end, window2.token2word_end),
        )

    @staticmethod
    def _merge_span_annotation(
        span_annotation1: List[list], span_annotation2: List[list]
    ) -> List[list]:
        uniq_store = set()
        final_span_annotation_store = []
        for span_annotation in itertools.chain(span_annotation1, span_annotation2):
            span_annotation_id = tuple(span_annotation)
            if span_annotation_id not in uniq_store:
                uniq_store.add(span_annotation_id)
                final_span_annotation_store.append(span_annotation)
        return sorted(final_span_annotation_store, key=lambda x: x[0])

    @staticmethod
    def _merge_predictions(
        window1: RelikReaderSample, window2: RelikReaderSample
    ) -> Tuple[Set[Tuple[int, int, str]], dict]:
        # a RelikReaderSample should have a filed called `predicted_spans`
        # that stores the span-level predictions, or a filed called
        # `predicted_triplets` that stores the triple-level predictions

        # span predictions
        merged_span_predictions: Set = set()
        merged_span_probabilities = dict()
        # triple predictions
        merged_triplet_predictions: Set = set()
        merged_triplet_probs: Dict = dict()

        if (
            getattr(window1, "predicted_spans", None) is not None
            and getattr(window2, "predicted_spans", None) is not None
        ):
            merged_span_predictions = set(window1.predicted_spans).union(
                set(window2.predicted_spans)
            )
            merged_span_predictions = sorted(merged_span_predictions)
            # probabilities
            for span_prediction, predicted_probs in itertools.chain(
                (
                    window1.probs_window_labels_chars.items()
                    if window1.probs_window_labels_chars is not None
                    else []
                ),
                (
                    window2.probs_window_labels_chars.items()
                    if window2.probs_window_labels_chars is not None
                    else []
                ),
            ):
                if span_prediction not in merged_span_probabilities:
                    merged_span_probabilities[span_prediction] = predicted_probs

            if (
                getattr(window1, "predicted_triplets", None) is not None
                and getattr(window2, "predicted_triplets", None) is not None
            ):
                # try to merge the triplets predictions
                # add offset to the second window
                window1_triplets = [
                    (
                        merged_span_predictions.index(window1.predicted_spans[t[0]]),
                        t[1],
                        merged_span_predictions.index(window1.predicted_spans[t[2]]),
                        t[3],
                    )
                    for t in window1.predicted_triplets
                ]
                window2_triplets = [
                    (
                        merged_span_predictions.index(window2.predicted_spans[t[0]]),
                        t[1],
                        merged_span_predictions.index(window2.predicted_spans[t[2]]),
                        t[3],
                    )
                    for t in window2.predicted_triplets
                ]
                merged_triplet_predictions = set(window1_triplets).union(
                    set(window2_triplets)
                )
                merged_triplet_predictions = sorted(merged_triplet_predictions)
                # for now no triplet probs, we don't need them for the moment

        return (
            merged_span_predictions,
            merged_span_probabilities,
            merged_triplet_predictions,
            merged_triplet_probs,
        )

    @staticmethod
    def _merge_candidates(window1: RelikReaderSample, window2: RelikReaderSample):
        # this is a list of lists of Documents, one for each window
        span_candidates = []
        if (
            getattr(window1, "span_candidates", None) is not None
            and len(window1.span_candidates) > 0
        ):
            # if this is already a merged window, then we have a list of lists
            if isinstance(window1.span_candidates[0], list):
                span_candidates = window1.span_candidates
            # otherwise we have directly a list of Documents
            else:
                span_candidates.append(window1.span_candidates)
        # now we can just add the second window candidates
        if getattr(window2, "span_candidates", None) is not None:
            span_candidates.append(window2.span_candidates)

        triplet_candidates = []
        if (
            getattr(window1, "triplet_candidates", None) is not None
            and len(window1.triplet_candidates) > 0
        ):
            if isinstance(window1.triplet_candidates[0], list):
                triplet_candidates = window1.triplet_candidates
            else:
                triplet_candidates.append(window1.triplet_candidates)
            triplet_candidates = window1.triplet_candidates
        if getattr(window2, "triplet_candidates", None) is not None:
            # triplet_candidates += window2.triplet_candidates
            triplet_candidates.append(window2.triplet_candidates)

        return span_candidates, triplet_candidates

    def _merge_window_pair(
        self,
        window1: RelikReaderSample,
        window2: RelikReaderSample,
    ) -> RelikReaderSample:
        merging_output = dict()

        if getattr(window1, "doc_id", None) is not None:
            assert window1.doc_id == window2.doc_id

        if getattr(window1, "offset", None) is not None:
            assert (
                window1.offset < window2.offset
            ), f"window 2 offset ({window2.offset}) is smaller that window 1 offset({window1.offset})"

        merging_output["doc_id"] = window1.doc_id
        merging_output["offset"] = window2.offset

        # We no longer merge these fields, as they are not used in the current implementation and caused lag
        # m_tokens, m_token2char_start, m_token2char_end = self._merge_tokens(
        #     window1, window2
        # )

        # m_words, m_token2word_start, m_token2word_end = self._merge_words(
        #     window1, window2
        # )

        (m_span_candidates, m_triplet_candidates) = self._merge_candidates(
            window1, window2
        )

        window_labels = None
        if getattr(window1, "window_labels", None) is not None:
            window_labels = self._merge_span_annotation(
                window1.window_labels, window2.window_labels
            )

        (
            predicted_spans,
            predicted_spans_probs,
            predicted_triplets,
            predicted_triplets_probs,
        ) = self._merge_predictions(window1, window2)

        # merge text, take into account overlapping chars
        if isinstance(self.splitter, SpacySentenceSplitter):
            m_text = window1.text[: window2.offset] + " " + window2.text
        else:
            m_text = window1.text[: window2.offset] + window2.text

        merging_output.update(
            dict(
                text=m_text,
                window_labels=window_labels,
                span_candidates=m_span_candidates,
                triplet_candidates=m_triplet_candidates,
                predicted_spans=predicted_spans,
                predicted_spans_probs=predicted_spans_probs,
                predicted_triplets=predicted_triplets,
                predicted_triplets_probs=predicted_triplets_probs,
            )
        )

        return RelikReaderSample(**merging_output)



import torch
from torch.optim.lr_scheduler import LRScheduler


class LinearSchedulerWithWarmup(LRScheduler):
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        num_warmup_steps: int,
        num_training_steps: int,
        last_epoch: int = -1,
        verbose: bool = False,
        **kwargs,
    ):
        self.num_warmup_steps = num_warmup_steps
        self.num_training_steps = num_training_steps
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        def scheduler_fn(current_step):
            if current_step < self.num_warmup_steps:
                return current_step / max(1, self.num_warmup_steps)
            return max(
                0.0,
                float(self.num_training_steps - current_step)
                / float(max(1, self.num_training_steps - self.num_warmup_steps)),
            )

        return [base_lr * scheduler_fn(self.last_epoch) for base_lr in self.base_lrs]


class LinearScheduler(LRScheduler):
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        num_training_steps: int,
        last_epoch: int = -1,
        verbose: bool = False,
        **kwargs,
    ):
        self.num_training_steps = num_training_steps
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        def scheduler_fn(current_step):
            # if current_step < self.num_warmup_steps:
            #     return current_step / max(1, self.num_warmup_steps)
            return max(
                0.0,
                float(self.num_training_steps - current_step)
                / float(max(1, self.num_training_steps)),
            )

        return [base_lr * scheduler_fn(self.last_epoch) for base_lr in self.base_lrs]



SPACY_LANGUAGE_MAPPER = {
    "ca": "ca_core_news_sm",
    "da": "da_core_news_sm",
    "de": "de_core_news_sm",
    "el": "el_core_news_sm",
    "en": "en_core_web_sm",
    "es": "es_core_news_sm",
    "fr": "fr_core_news_sm",
    "it": "it_core_news_sm",
    "ja": "ja_core_news_sm",
    "lt": "lt_core_news_sm",
    "mk": "mk_core_news_sm",
    "nb": "nb_core_news_sm",
    "nl": "nl_core_news_sm",
    "pl": "pl_core_news_sm",
    "pt": "pt_core_news_sm",
    "ro": "ro_core_news_sm",
    "ru": "ru_core_news_sm",
    "xx": "xx_sent_ud_sm",
    "zh": "zh_core_web_sm",
    "ca_core_news_sm": "ca_core_news_sm",
    "ca_core_news_md": "ca_core_news_md",
    "ca_core_news_lg": "ca_core_news_lg",
    "ca_core_news_trf": "ca_core_news_trf",
    "da_core_news_sm": "da_core_news_sm",
    "da_core_news_md": "da_core_news_md",
    "da_core_news_lg": "da_core_news_lg",
    "da_core_news_trf": "da_core_news_trf",
    "de_core_news_sm": "de_core_news_sm",
    "de_core_news_md": "de_core_news_md",
    "de_core_news_lg": "de_core_news_lg",
    "de_dep_news_trf": "de_dep_news_trf",
    "el_core_news_sm": "el_core_news_sm",
    "el_core_news_md": "el_core_news_md",
    "el_core_news_lg": "el_core_news_lg",
    "en_core_web_sm": "en_core_web_sm",
    "en_core_web_md": "en_core_web_md",
    "en_core_web_lg": "en_core_web_lg",
    "en_core_web_trf": "en_core_web_trf",
    "es_core_news_sm": "es_core_news_sm",
    "es_core_news_md": "es_core_news_md",
    "es_core_news_lg": "es_core_news_lg",
    "es_dep_news_trf": "es_dep_news_trf",
    "fr_core_news_sm": "fr_core_news_sm",
    "fr_core_news_md": "fr_core_news_md",
    "fr_core_news_lg": "fr_core_news_lg",
    "fr_dep_news_trf": "fr_dep_news_trf",
    "it_core_news_sm": "it_core_news_sm",
    "it_core_news_md": "it_core_news_md",
    "it_core_news_lg": "it_core_news_lg",
    "ja_core_news_sm": "ja_core_news_sm",
    "ja_core_news_md": "ja_core_news_md",
    "ja_core_news_lg": "ja_core_news_lg",
    "ja_dep_news_trf": "ja_dep_news_trf",
    "lt_core_news_sm": "lt_core_news_sm",
    "lt_core_news_md": "lt_core_news_md",
    "lt_core_news_lg": "lt_core_news_lg",
    "mk_core_news_sm": "mk_core_news_sm",
    "mk_core_news_md": "mk_core_news_md",
    "mk_core_news_lg": "mk_core_news_lg",
    "nb_core_news_sm": "nb_core_news_sm",
    "nb_core_news_md": "nb_core_news_md",
    "nb_core_news_lg": "nb_core_news_lg",
    "nl_core_news_sm": "nl_core_news_sm",
    "nl_core_news_md": "nl_core_news_md",
    "nl_core_news_lg": "nl_core_news_lg",
    "pl_core_news_sm": "pl_core_news_sm",
    "pl_core_news_md": "pl_core_news_md",
    "pl_core_news_lg": "pl_core_news_lg",
    "pt_core_news_sm": "pt_core_news_sm",
    "pt_core_news_md": "pt_core_news_md",
    "pt_core_news_lg": "pt_core_news_lg",
    "ro_core_news_sm": "ro_core_news_sm",
    "ro_core_news_md": "ro_core_news_md",
    "ro_core_news_lg": "ro_core_news_lg",
    "ru_core_news_sm": "ru_core_news_sm",
    "ru_core_news_md": "ru_core_news_md",
    "ru_core_news_lg": "ru_core_news_lg",
    "xx_ent_wiki_sm": "xx_ent_wiki_sm",
    "xx_sent_ud_sm": "xx_sent_ud_sm",
    "zh_core_web_sm": "zh_core_web_sm",
    "zh_core_web_md": "zh_core_web_md",
    "zh_core_web_lg": "zh_core_web_lg",
    "zh_core_web_trf": "zh_core_web_trf",
}

from relik.inference.data.tokenizers.spacy_tokenizer import SpacyTokenizer



import argparse
import os
from typing import Tuple

import omegaconf
import torch

from relik.common.utils import from_cache
from relik.reader.lightning_modules.relik_reader_pl_module import RelikReaderPLModule

CKPT_FILE_NAME = "model.ckpt"
CONFIG_FILE_NAME = "cfg.yaml"


def convert_pl_module(pl_module_ckpt_path: str, output_dir: str) -> None:
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    else:
        print(f"{output_dir} already exists, aborting operation")
        exit(1)

    relik_pl_module: RelikReaderPLModule = RelikReaderPLModule.load_from_checkpoint(
        pl_module_ckpt_path
    )
    torch.save(
        relik_pl_module.relik_reader_core_model, f"{output_dir}/{CKPT_FILE_NAME}"
    )
    with open(f"{output_dir}/{CONFIG_FILE_NAME}", "w") as f:
        omegaconf.OmegaConf.save(
            omegaconf.OmegaConf.create(relik_pl_module.hparams["cfg"]), f
        )


def load_model_and_conf(
    model_dir_path: str,
) -> Tuple[torch.nn.Module, omegaconf.DictConfig]:
    # TODO: quick workaround to load the model from HF hub
    model_dir = from_cache(
        model_dir_path,
        filenames=[CKPT_FILE_NAME, CONFIG_FILE_NAME],
        cache_dir=None,
        force_download=False,
    )

    ckpt_path = f"{model_dir}/{CKPT_FILE_NAME}"
    model = torch.load(ckpt_path, map_location=torch.device("cpu"))

    model_cfg_path = f"{model_dir}/{CONFIG_FILE_NAME}"
    model_conf = omegaconf.OmegaConf.load(model_cfg_path)
    return model, model_conf


def parse_arg() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ckpt",
        help="Path to the pytorch lightning ckpt you want to convert.",
        required=True,
    )
    parser.add_argument(
        "--output-dir",
        "-o",
        help="The output dir to store the bare models and the config.",
        required=True,
    )
    return parser.parse_args()


def main():
    args = parse_arg()
    convert_pl_module(args.ckpt, args.output_dir)


if __name__ == "__main__":
    main()



"""
Simple check list from AllenNLP repo: https://github.com/allenai/allennlp/blob/master/setup.py
To create the package for pypi.
1. Run `make pre-release` (or `make pre-patch` for a patch release) then run `make fix-copies` to fix the index of the
   documentation.
2. Run Tests for Amazon Sagemaker. The documentation is located in `./tests/sagemaker/README.md`, otherwise @philschmid.
3. Unpin specific versions from setup.py that use a git install.
4. Commit these changes with the message: "Release: VERSION"
5. Add a tag in git to mark the release: "git tag VERSION -m 'Adds tag VERSION for pypi' "
   Push the tag to git: git push --tags origin master
6. Build both the sources and the wheel. Do not change anything in setup.py between
   creating the wheel and the source distribution (obviously).
   For the wheel, run: "python setup.py bdist_wheel" in the top level directory.
   (this will build a wheel for the python version you use to build it).
   For the sources, run: "python setup.py sdist"
   You should now have a /dist directory with both .whl and .tar.gz source versions.
7. Check that everything looks correct by uploading the package to the pypi test server:
   twine upload dist/* -r pypitest
   (pypi suggest using twine as other methods upload files via plaintext.)
   You may have to specify the repository url, use the following command then:
   twine upload dist/* -r pypitest --repository-url=https://test.pypi.org/legacy/
   Check that you can install it in a virtualenv by running:
   pip install -i https://testpypi.python.org/pypi transformers
8. Upload the final version to actual pypi:
   twine upload dist/* -r pypi
9. Copy the release notes from RELEASE.md to the tag in github once everything is looking hunky-dory.
10. Run `make post-release` (or `make post-patch` for a patch release).
"""
from collections import defaultdict

import setuptools


def parse_requirements_file(
    path, allowed_extras: set = None, include_all_extra: bool = True
):
    requirements = []
    extras = defaultdict(list)
    find_links = []
    with open(path) as requirements_file:
        import re

        def fix_url_dependencies(req: str) -> str:
            """Pip and setuptools disagree about how URL dependencies should be handled."""
            m = re.match(
                r"^(git\+)?(https|ssh)://(git@)?github\.com/([\w-]+)/(?P<name>[\w-]+)\.git",
                req,
            )
            if m is None:
                return req
            else:
                return f"{m.group('name')} @ {req}"

        for line in requirements_file:
            line = line.strip()
            if line.startswith("#") or len(line) <= 0:
                continue
            if (
                line.startswith("-f")
                or line.startswith("--find-links")
                or line.startswith("--index-url")
                or line.startswith("--extra-index-url")
            ):
                find_links.append(line.split(" ", maxsplit=1)[-1].strip())
                continue

            req, *needed_by = line.split("# needed by:")
            req = fix_url_dependencies(req.strip())
            if needed_by:
                for extra in needed_by[0].strip().split(","):
                    extra = extra.strip()
                    if allowed_extras is not None and extra not in allowed_extras:
                        raise ValueError(f"invalid extra '{extra}' in {path}")
                    extras[extra].append(req)
                if include_all_extra and req not in extras["all"]:
                    # if "gpu" in extra:
                    #     extras["all-gpu"].append(req)
                    # else:
                    extras["all"].append(req)

            else:
                requirements.append(req)
    return requirements, extras, find_links


allowed_extras = {
    "serve",
    "ray",
    "train",
    "retriever",
    "reader",
    "all",
    "faiss",
    "faiss-gpu",
    "dev",
}

# Load requirements.
install_requirements, extras, find_links = parse_requirements_file(
    "requirements.txt", allowed_extras=allowed_extras
)

# version.py defines the VERSION and VERSION_SHORT variables.
# We use exec here, so we don't import allennlp whilst setting up.
VERSION = {}  # type: ignore
with open("relik/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="relik",
    version=VERSION["VERSION"],
    author="Edoardo Barba, Riccardo Orlando, Pere-Lluís Huguet Cabot",
    author_email="orlandorcc@gmail.com",
    description="Fast and Accurate Entity Linking and Relation Extraction on an Academic Budget",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SapienzaNLP/relik",
    keywords="NLP Sapienza sapienzanlp deep learning transformer pytorch retriever entity linking relation extraction reader budget",
    packages=setuptools.find_packages(),
    include_package_data=True,
    license="Apache",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    install_requires=install_requirements,
    extras_require=extras,
    python_requires=">=3.10",
    find_links=find_links,
    entry_points={
        "console_scripts": ["relik = relik.cli.cli:app"],
    },
)



import logging
import os
import sys
import threading
from logging.config import dictConfig
from typing import Any, Dict, Optional

from art import text2art, tprint
from colorama import Fore, Style, init
from rich import get_console
from termcolor import colored, cprint


_lock = threading.Lock()
_default_handler: Optional[logging.Handler] = None

_default_log_level = logging.WARNING

# fancy logger
_console = get_console()


class ColorfulFormatter(logging.Formatter):
    """
    Formatter to add coloring to log messages by log type
    """

    COLORS = {
        "WARNING": Fore.YELLOW,
        "ERROR": Fore.RED,
        "CRITICAL": Fore.RED + Style.BRIGHT,
        "DEBUG": Fore.CYAN,
        # "INFO": Fore.GREEN,
    }

    def format(self, record):
        record.rank = int(os.getenv("LOCAL_RANK", "0"))
        log_message = super().format(record)
        return self.COLORS.get(record.levelname, "") + log_message + Fore.RESET


DEFAULT_LOGGING_CONFIG: Dict[str, Any] = {
    "version": 1,
    "formatters": {
        "simple": {
            "format": "[%(asctime)s] [%(levelname)s] [%(name)s.%(funcName)s:%(lineno)d] [PID:%(process)d] %(message)s",
        },
        "colorful": {
            "()": ColorfulFormatter,
            "format": "[%(asctime)s] [%(levelname)s] [%(name)s.%(funcName)s:%(lineno)d] [PID:%(process)d] [RANK:%(rank)d] %(message)s",
        },
    },
    "filters": {},
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "filters": [],
            "stream": sys.stdout,
        },
        "color_console": {
            "class": "logging.StreamHandler",
            "formatter": "colorful",
            "filters": [],
            "stream": sys.stdout,
        },
    },
    "root": {"handlers": ["console"], "level": os.getenv("LOG_LEVEL", "INFO")},
    "loggers": {
        "relik": {
            "handlers": ["color_console"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
}


def configure_logging(**kwargs):
    """Configure with default logging"""
    init()  # Initialize colorama
    # merge DEFAULT_LOGGING_CONFIG with kwargs
    logger_config = DEFAULT_LOGGING_CONFIG
    if kwargs:
        logger_config.update(kwargs)
    dictConfig(logger_config)


def _get_library_name() -> str:
    return __name__.split(".")[0]


def _get_library_root_logger() -> logging.Logger:
    return logging.getLogger(_get_library_name())


def _configure_library_root_logger() -> None:
    global _default_handler

    with _lock:
        if _default_handler:
            # This library has already configured the library root logger.
            return
        _default_handler = logging.StreamHandler()  # Set sys.stderr as stream.
        _default_handler.flush = sys.stderr.flush

        # Apply our default configuration to the library root logger.
        library_root_logger = _get_library_root_logger()
        library_root_logger.addHandler(_default_handler)
        library_root_logger.setLevel(_default_log_level)
        library_root_logger.propagate = False


def _reset_library_root_logger() -> None:
    global _default_handler

    with _lock:
        if not _default_handler:
            return

        library_root_logger = _get_library_root_logger()
        library_root_logger.removeHandler(_default_handler)
        library_root_logger.setLevel(logging.NOTSET)
        _default_handler = None


def set_log_level(level: int, logger: logging.Logger = None) -> None:
    """
    Set the log level.
    Args:
        level (:obj:`int`):
            Logging level.
        logger (:obj:`logging.Logger`):
            Logger to set the log level.
    """
    if not logger:
        _configure_library_root_logger()
        logger = _get_library_root_logger()
    logger.setLevel(level)


def get_logger(
    name: Optional[str] = None,
    level: Optional[int] = None,
    formatter: Optional[str] = None,
    **kwargs,
) -> logging.Logger:
    """
    Return a logger with the specified name.
    """

    configure_logging(**kwargs)

    if name is None:
        name = _get_library_name()

    _configure_library_root_logger()

    if level is not None:
        set_log_level(level)

    if formatter is None:
        formatter = logging.Formatter(
            "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
        )
    _default_handler.setFormatter(formatter)

    return logging.getLogger(name)


def get_console_logger():
    return _console


def print_relik_text_art(
    text: str = "relik", font: str = "larry3d", color: str = "magenta", **kwargs
):
    # tprint(text, font=font, **kwargs)
    art = text2art(text, font=font, **kwargs)  # .rstrip()
    # art += "\n\n           Retrieve, Read, and Link"
    # art += "\nA fast and lightweight Information Extraction framework"
    cprint(art, color, attrs=["bold"])



# from relik.reader.pytorch_modules.hf.modeling_relik import RelikReaderSpanModel
# from relik.reader.pytorch_modules.span import RelikReaderForSpanExtraction
# from relik.reader.pytorch_modules.triplet import RelikReaderForTripletExtraction


RELIK_READER_CLASS_MAP = {
    "RelikReaderSpanModel": "relik.reader.pytorch_modules.span.RelikReaderForSpanExtraction",
    "RelikReaderREModel": "relik.reader.pytorch_modules.triplet.RelikReaderForTripletExtraction",
}



from typing import List

import torch
import transformers
from torch.optim import AdamW


class AdamWWithWarmupOptimizer:
    def __init__(
        self,
        lr: float | List[float],
        warmup_steps: int,
        total_steps: int,
        weight_decay: float,
        no_decay_params: List[str],
        other_lr_params: List[str] = None,
    ):
        self.lr = lr[0] if isinstance(lr, list) else lr
        self.lr2 = lr[1] if isinstance(lr, list) else lr
        self.warmup_steps = warmup_steps
        self.total_steps = total_steps
        self.weight_decay = weight_decay
        self.no_decay_params = no_decay_params
        self.other_lr_params = other_lr_params or []  # Ensure it's a list

    def group_params(self, module: torch.nn.Module) -> list:
        decay_params = set()
        no_decay_params = set()
        other_lr_params = set()
        # Populate parameter sets
        for n, p in module.named_parameters():
            if any(nd in n for nd in self.no_decay_params):
                no_decay_params.add(p)
            elif any(olr in n for olr in self.other_lr_params):
                other_lr_params.add(p)
            else:
                decay_params.add(p)
        # Group parameters
        optimizer_grouped_parameters = []
        if decay_params:
            optimizer_grouped_parameters.append(
                {"params": list(decay_params), "weight_decay": self.weight_decay}
            )
        if no_decay_params:
            optimizer_grouped_parameters.append(
                {"params": list(no_decay_params), "weight_decay": 0.0}
            )
        if other_lr_params:
            optimizer_grouped_parameters.append(
                {
                    "params": list(other_lr_params),
                    "lr": self.lr2,
                    "weight_decay": self.weight_decay,
                }
            )

        return optimizer_grouped_parameters

    def __call__(self, module: torch.nn.Module):
        optimizer_grouped_parameters = self.group_params(module)
        optimizer = AdamW(
            optimizer_grouped_parameters, lr=self.lr, weight_decay=self.weight_decay
        )
        scheduler = transformers.get_linear_schedule_with_warmup(
            optimizer, self.warmup_steps, self.total_steps
        )
        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": scheduler,
                "interval": "step",
                "frequency": 1,
            },
        }



from dataclasses import dataclass

import torch

from relik.retriever.indexers.document import Document

PRECISION_MAP = {
    None: torch.float32,
    32: torch.float32,
    16: torch.float16,
    torch.float32: torch.float32,
    torch.float16: torch.float16,
    torch.bfloat16: torch.bfloat16,
    "float32": torch.float32,
    "float16": torch.float16,
    "bfloat16": torch.bfloat16,
    "float": torch.float32,
    "half": torch.float16,
    "32": torch.float32,
    "16": torch.float16,
    "fp32": torch.float32,
    "fp16": torch.float16,
    "bf16": torch.bfloat16,
}


@dataclass
class RetrievedSample:
    """
    Dataclass for the output of the GoldenRetriever model.
    """

    score: float
    document: Document



import logging
import os
from pathlib import Path
from typing import Any, Dict, List

import torch
import transformers as tr
from torch.utils.data import IterableDataset
from transformers import AutoConfig, AutoModel

from relik.common.log import get_logger

from relik.common.utils import get_callable_from_string
from relik.inference.data.objects import AnnotationType

# from relik.reader.pytorch_modules import RELIK_READER_CLASS_MAP
from relik.reader.pytorch_modules import RELIK_READER_CLASS_MAP
from relik.reader.pytorch_modules.hf.modeling_relik import (
    RelikReaderConfig,
    # RelikReaderREModel,
    RelikReaderSample,
    # RelikReaderSpanModel,
)

# from relik.reader.pytorch_modules.span import RelikReaderForSpanExtraction
# from relik.reader.pytorch_modules.triplet import RelikReaderForTripletExtraction
from relik.retriever.pytorch_modules import PRECISION_MAP

logger = get_logger(__name__, level=logging.INFO)


class RelikReaderBase(torch.nn.Module):
    default_reader_class: str | None = None
    default_data_class: str | None = None

    def __init__(
        self,
        transformer_model: str | tr.PreTrainedModel | None = None,
        additional_special_symbols: int = 0,
        num_layers: int | None = None,
        activation: str = "gelu",
        linears_hidden_size: int | None = 512,
        use_last_k_layers: int = 1,
        training: bool = False,
        device: str | torch.device | None = None,
        precision: int = 32,
        tokenizer: str | tr.PreTrainedTokenizer | None = None,
        dataset: IterableDataset | str | None = None,
        default_reader_class: tr.PreTrainedModel | str | None = None,
        **kwargs,
    ) -> None:
        super().__init__()

        self.default_reader_class = default_reader_class or self.default_reader_class

        if self.default_reader_class is None:
            raise ValueError("You must specify a default reader class.")

        # get the callable for the default reader class
        self.default_reader_class: tr.PreTrainedModel = get_callable_from_string(
            self.default_reader_class
        )

        if isinstance(transformer_model, str):
            self.name_or_path = transformer_model
            config = AutoConfig.from_pretrained(
                transformer_model, trust_remote_code=True
            )
            if "relik-reader" in config.model_type:
                transformer_model = self.default_reader_class.from_pretrained(
                    transformer_model,
                    config=config,
                    ignore_mismatched_sizes=True,
                    trust_remote_code=True,
                    **kwargs,
                )
            else:
                reader_config = RelikReaderConfig(
                    transformer_model=transformer_model,
                    additional_special_symbols=additional_special_symbols,
                    num_layers=num_layers,
                    activation=activation,
                    linears_hidden_size=linears_hidden_size,
                    use_last_k_layers=use_last_k_layers,
                    training=training,
                    **kwargs,
                )
                transformer_model = self.default_reader_class(reader_config)
                self.name_or_path = transformer_model.config.transformer_model

        self.relik_reader_model = transformer_model

        self.relik_reader_model_config = self.relik_reader_model.config
        # self.name_or_path = self.relik_reader_model_config.name_or_path
        # self.name_or_path = self.relik_reader_model.config.transformer_model

        # get the tokenizer
        self._tokenizer = tokenizer

        # and instantiate the dataset class
        self.dataset: IterableDataset | None = dataset

        # move the model to the device
        self.to(device or torch.device("cpu"))

        # set the precision
        self.precision = precision
        self.to(PRECISION_MAP[precision])

    def forward(self, **kwargs) -> Dict[str, Any]:
        return self.relik_reader_model(**kwargs)

    def _read(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    @torch.no_grad()
    @torch.inference_mode()
    def read(
        self,
        text: List[str] | List[List[str]] | None = None,
        samples: List[RelikReaderSample] | None = None,
        input_ids: torch.Tensor | None = None,
        attention_mask: torch.Tensor | None = None,
        token_type_ids: torch.Tensor | None = None,
        prediction_mask: torch.Tensor | None = None,
        special_symbols_mask: torch.Tensor | None = None,
        candidates: List[List[str]] | None = None,
        max_length: int = 1000,
        max_batch_size: int = 128,
        token_batch_size: int = 2048,
        precision: int | str | None = None,
        annotation_type: str | AnnotationType = AnnotationType.CHAR,
        progress_bar: bool = False,
        *args,
        **kwargs,
    ) -> List[RelikReaderSample] | List[List[RelikReaderSample]]:
        """
        Reads the given text.

        Args:
            text (:obj:`List[str]` or :obj:`List[List[str]]`, `optional`):
                The text to read in tokens. If a list of list of tokens is provided, each
                inner list is considered a sentence.
            samples (:obj:`List[RelikReaderSample]`, `optional`):
                The samples to read. If provided, `text` and `candidates` are ignored.
            input_ids (:obj:`torch.Tensor`, `optional`):
                The input ids of the text.
            attention_mask (:obj:`torch.Tensor`, `optional`):
                The attention mask of the text.
            token_type_ids (:obj:`torch.Tensor`, `optional`):
                The token type ids of the text.
            prediction_mask (:obj:`torch.Tensor`, `optional`):
                The prediction mask of the text.
            special_symbols_mask (:obj:`torch.Tensor`, `optional`):
                The special symbols mask of the text.
            candidates (:obj:`List[List[str]]`, `optional`):
                The candidates of the text.
            max_length (:obj:`int`, `optional`, defaults to 1024):
                The maximum length of the text.
            max_batch_size (:obj:`int`, `optional`, defaults to 128):
                The maximum batch size.
            token_batch_size (:obj:`int`, `optional`):
                The maximum number of tokens per batch.
            precision (:obj:`int` or :obj:`str`, `optional`):
                The precision to use. If not provided, the default is 32 bit.
            annotation_type (`str` or `AnnotationType`, `optional`, defaults to `char`):
                The type of annotation to return. If `char`, the spans will be in terms of
                character offsets. If `word`, the spans will be in terms of word offsets.
            progress_bar (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether to show a progress bar.

        Returns:
            The predicted labels for each sample.
        """
        if isinstance(annotation_type, str):
            try:
                annotation_type = AnnotationType(annotation_type)
            except ValueError:
                raise ValueError(
                    f"Annotation type `{annotation_type}` not recognized. "
                    f"Please choose one of {list(AnnotationType)}."
                )

        if text is None and input_ids is None and samples is None:
            raise ValueError(
                "Either `text` or `input_ids` or `samples` must be provided."
            )
        if (input_ids is None and samples is None) and (
            text is None or candidates is None
        ):
            raise ValueError(
                "`text` and `candidates` must be provided to return the predictions when "
                "`input_ids` and `samples` is not provided."
            )
        if text is not None and samples is None:
            if len(text) != len(candidates):
                raise ValueError("`text` and `candidates` must have the same length.")
            if isinstance(text[0], str):  # change to list of text
                text = [text]
                candidates = [candidates]

            samples = [
                RelikReaderSample(tokens=t, candidates=c)
                for t, c in zip(text, candidates)
            ]

        return self._read(
            samples,
            input_ids,
            attention_mask,
            token_type_ids,
            prediction_mask,
            special_symbols_mask,
            max_length,
            max_batch_size,
            token_batch_size,
            precision or self.precision,
            annotation_type,
            progress_bar,
            *args,
            **kwargs,
        )

    @property
    def device(self) -> torch.device:
        """
        The device of the model.
        """
        return next(self.parameters()).device

    @property
    def tokenizer(self) -> tr.PreTrainedTokenizer:
        """
        The tokenizer.
        """
        if self._tokenizer:
            return self._tokenizer

        self._tokenizer = tr.AutoTokenizer.from_pretrained(
            self.relik_reader_model.config.name_or_path
            if self.relik_reader_model.config.name_or_path
            else self.relik_reader_model.config.transformer_model
        )
        return self._tokenizer

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_dir: str | os.PathLike,
        **kwargs,
    ):
        transformer_model = AutoModel.from_pretrained(
            model_name_or_dir, trust_remote_code=True, **kwargs
        )
        if transformer_model.__class__.__name__ not in RELIK_READER_CLASS_MAP:
            raise ValueError(f"Model type {type(transformer_model)} not recognized.")

        reader_class = RELIK_READER_CLASS_MAP[transformer_model.__class__.__name__]
        reader_class = get_callable_from_string(reader_class)
        return reader_class(transformer_model=transformer_model, **kwargs)

    def save_pretrained(
        self,
        output_dir: str | os.PathLike,
        model_name: str | None = None,
        push_to_hub: bool = False,
        **kwargs,
    ) -> None:
        """
        Saves the model to the given path.

        Args:
            output_dir (`str` or :obj:`os.PathLike`):
                The path to save the model to.
            model_name (`str`, `optional`):
                The name of the model. If not provided, the model will be saved as
                `default_reader_class.__name__`.
            push_to_hub (`bool`, `optional`, defaults to `False`):
                Whether to push the model to the HuggingFace Hub.
            **kwargs:
                Additional keyword arguments to pass to the `save_pretrained` method
        """
        # create the output directory
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        model_name = model_name or output_dir.name

        logger.info(f"Saving reader to {output_dir / model_name}")

        # save the model
        self.relik_reader_model.config.register_for_auto_class()
        self.relik_reader_model.register_for_auto_class()
        self.relik_reader_model.save_pretrained(
            str(output_dir / model_name), push_to_hub=push_to_hub, **kwargs
        )

        if self.tokenizer:
            logger.info("Saving also the tokenizer")
            self.tokenizer.save_pretrained(
                str(output_dir / model_name), push_to_hub=push_to_hub, **kwargs
            )



import os
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import torch
from torch.utils.data import Dataset, IterableDataset

from relik.common.log import get_logger

logger = get_logger(__name__)


class BaseDataset(Dataset):
    def __init__(
        self,
        name: str,
        path: Optional[Union[str, os.PathLike, List[str], List[os.PathLike]]] = None,
        data: Any = None,
        **kwargs,
    ):
        super().__init__()
        self.name = name
        if path is None and data is None:
            raise ValueError("Either `path` or `data` must be provided")
        self.path = path
        self.project_folder = Path(__file__).parent.parent.parent
        self.data = data

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(
        self, index
    ) -> Union[Dict[str, torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        return self.data[index]

    def __repr__(self) -> str:
        return f"Dataset({self.name=}, {self.path=})"

    def load(
        self,
        paths: Union[str, os.PathLike, List[str], List[os.PathLike]],
        *args,
        **kwargs,
    ) -> Any:
        # load data from single or multiple paths in one single dataset
        raise NotImplementedError

    @staticmethod
    def collate_fn(batch: Any, *args, **kwargs) -> Any:
        raise NotImplementedError


class IterableBaseDataset(IterableDataset):
    def __init__(
        self,
        name: str,
        path: Optional[Union[str, Path, List[str], List[Path]]] = None,
        data: Any = None,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.name = name
        if path is None and data is None:
            raise ValueError("Either `path` or `data` must be provided")
        self.path = path
        self.project_folder = Path(__file__).parent.parent.parent
        self.data = data

    def __iter__(self) -> Iterator[Dict[str, torch.Tensor]]:
        for sample in self.data:
            yield sample

    def __repr__(self) -> str:
        return f"Dataset({self.name=}, {self.path=})"

    def load(
        self,
        paths: Union[str, os.PathLike, List[str], List[os.PathLike]],
        *args,
        **kwargs,
    ) -> Any:
        # load data from single or multiple paths in one single dataset
        raise NotImplementedError

    @staticmethod
    def collate_fn(batch: Any, *args, **kwargs) -> Any:
        raise NotImplementedError



from typing import Any, Dict

import hydra
import torch
from omegaconf import DictConfig, OmegaConf

from relik.common.log import get_logger
from relik.inference.data.objects import TaskType
from relik.reader.pytorch_modules.base import RelikReaderBase
from relik.retriever.indexers.base import BaseDocumentIndex
from relik.retriever.pytorch_modules import PRECISION_MAP
from relik.retriever.pytorch_modules.model import GoldenRetriever

logger = get_logger(__name__)


def _instantiate_retriever(
    retriever: GoldenRetriever | DictConfig | Dict,
    device: str | None,
    precision: int | str | torch.dtype | None,
    **kwargs: Any,
) -> GoldenRetriever:
    """
    Instantiate a retriever.

    Args:
        retriever (`GoldenRetriever`, `DictConfig` or `Dict`):
            The retriever to instantiate.
        retriever_device (`str`, `optional`):
            The device to use for the retriever.
        retriever_precision (`int`, `str` or `torch.dtype`, `optional`):
            The precision to use for the retriever.
        retriever_kwargs (`Dict[str, Any]`, `optional`):
            Additional keyword arguments to pass to the retriever.

    Returns:
        `GoldenRetriever`:
            The instantiated retriever.
    """
    if not isinstance(retriever, GoldenRetriever):
        # convert to DictConfig
        retriever = hydra.utils.instantiate(
            OmegaConf.create(retriever),
            device=device,
            precision=precision,
            **kwargs,
        )
    else:
        if device is not None:
            logger.info(f"Moving retriever to `{device}`.")
            retriever.to(device)
        if precision is not None:
            logger.info(
                f"Setting precision of retriever to `{PRECISION_MAP[precision]}`."
            )
            retriever.to(PRECISION_MAP[precision])
    retriever.training = False
    retriever.eval()
    return retriever


def load_retriever(
    retriever: GoldenRetriever | DictConfig | Dict | str,
    device: str | None | torch.device | int = None,
    precision: int | str | torch.dtype | None = None,
    task: TaskType | str | None = None,
    compile: bool = False,
    **kwargs,
) -> Dict[TaskType, GoldenRetriever]:
    """
    Load and instantiate retrievers for a given task.

    Args:
        retriever (GoldenRetriever | DictConfig | Dict):
            The retriever object or configuration.
        device (str | None):
            The device to load the retriever on.
        precision (int | str | torch.dtype | None):
            The precision of the retriever.
        task (TaskType):
            The task type for the retriever.
        compile (bool, optional):
            Whether to compile the retriever. Defaults to False.
        **kwargs:
            Additional keyword arguments to be passed to the retriever instantiation.

    Returns:
        Dict[TaskType, GoldenRetriever]:
            A dictionary containing the instantiated retrievers.

    Raises:
        ValueError: If the `retriever` argument is not of type `GoldenRetriever`, `DictConfig`, or `Dict`.
        ValueError: If the `retriever` argument is a `DictConfig` without the `_target_` key.
        ValueError: If the task type is not valid for each retriever in the `DictConfig`.

    """

    # retriever section
    _retriever: Dict[TaskType, GoldenRetriever] = {
        TaskType.SPAN: None,
        TaskType.TRIPLET: None,
    }

    # check retriever type, it can be a GoldenRetriever, a DictConfig or a Dict
    if not isinstance(retriever, (GoldenRetriever, DictConfig, Dict, str)):
        raise ValueError(
            f"`retriever` must be a `GoldenRetriever`, a `DictConfig`, "
            f"a `Dict`, or a `str`, got `{type(retriever)}`."
        )
    if isinstance(retriever, str):
        logger.warning(
            "Using a string to instantiate the retriever. "
            f"We will use the same model `{retriever}` for both query and passage encoder. "
            "If you want to use different models, please provide a dictionary with keys `question_encoder` and `passage_encoder`."
        )
        retriever = {
            "question_encoder": retriever,
            "_target_": "relik.retriever.pytorch_modules.model.GoldenRetriever",
        }
    # we need to check weather the DictConfig is a DictConfig for an instance of GoldenRetriever
    # or a primitive Dict
    # if isinstance(retriever, (DictConfig, Dict)):
    #     # then it is probably a primitive Dict
    #     if "_target_" not in retriever:
    #         retriever["_target_"] = "relik.retriever.pytorch_modules.model.GoldenRetriever"
    #         retriever = OmegaConf.to_container(retriever, resolve=True)
    #         # convert the key to TaskType
    #         try:
    #             retriever = {TaskType(k.lower()): v for k, v in retriever.items()}
    #         except ValueError as e:
    #             raise ValueError(
    #                 f"Please choose a valid task type (one of {list(TaskType)}) for each retriever."
    #             ) from e

    try:
        # convert the key to TaskType
        retriever = {TaskType(k): v for k, v in retriever.items()}
    except ValueError as e:
        retriever = {task: retriever}

    # instantiate each retriever
    if task in [TaskType.SPAN, TaskType.BOTH]:
        _retriever[TaskType.SPAN] = _instantiate_retriever(
            retriever[TaskType.SPAN],
            device,
            precision,
            **kwargs,
        )
    if task in [TaskType.TRIPLET, TaskType.BOTH]:
        _retriever[TaskType.TRIPLET] = _instantiate_retriever(
            retriever[TaskType.TRIPLET],
            device,
            precision,
            **kwargs,
        )

    # clean up None retrievers from the dictionary
    _retriever = {task_type: r for task_type, r in _retriever.items() if r is not None}
    if compile:
        # torch compile
        _retriever = {
            task_type: torch.compile(r) for task_type, r in _retriever.items()
        }

    return _retriever


def _instantiate_index(
    index: BaseDocumentIndex | DictConfig | Dict,
    device: str | None | torch.device | int = None,
    precision: int | str | torch.dtype | None = None,
    **kwargs: Dict[str, Any],
) -> BaseDocumentIndex:
    """
    Instantiate a document index.

    Args:
        index (`BaseDocumentIndex`, `DictConfig` or `Dict`):
            The document index to instantiate.
        device (`str`, `optional`):
            The device to use for the document index.
        precision (`int`, `str` or `torch.dtype`, `optional`):
            The precision to use for the document index.
        kwargs (`Dict[str, Any]`, `optional`):
            Additional keyword arguments to pass to the document index.

    Returns:
        `BaseDocumentIndex`:
            The instantiated document index.
    """
    if not isinstance(index, BaseDocumentIndex):
        index = OmegaConf.create(index)
        # use_faiss = kwargs.get("use_faiss", False)
        if "use_faiss" not in kwargs:
            kwargs["use_faiss"] = False
        # if use_faiss:
        #     index = OmegaConf.merge(
        #         index,
        #         {
        #             "_target_": "relik.retriever.indexers.faissindex.FaissDocumentIndex.from_pretrained",
        #         },
        #     )
        # else:
        # index = OmegaConf.merge(
        #     index,
        #     {
        #         "_target_": "relik.retriever.indexers.base.BaseDocumentIndex.from_pretrained",
        #     },
        # )
        if device is not None:
            kwargs["device"] = device
        if precision is not None:
            kwargs["precision"] = precision

        if "_target_" not in index:
            index["_target_"] = (
                "relik.retriever.indexers.base.BaseDocumentIndex.from_pretrained"
            )

        # merge the kwargs
        index = OmegaConf.merge(index, OmegaConf.create(kwargs))
        index: BaseDocumentIndex = hydra.utils.instantiate(index)
    else:
        index = index
        if device is not None:
            logger.info(f"Moving index to `{device}`.")
            index.to(device)
        if precision is not None:
            logger.info(f"Setting precision of index to `{PRECISION_MAP[precision]}`.")
            index.to(PRECISION_MAP[precision])
    return index


def load_index(
    index: BaseDocumentIndex | DictConfig | Dict | str,
    device: str | None,
    precision: int | str | torch.dtype | None,
    task: TaskType,
    **kwargs,
) -> Dict[TaskType, BaseDocumentIndex]:
    """
    Load the document index based on the specified parameters.

    Args:
        index (BaseDocumentIndex | DictConfig | Dict):
            The document index to load. It can be an instance of `BaseDocumentIndex`, a `DictConfig`, or a `Dict`.
        device (str | None):
            The device to use for loading the index. If `None`, the default device will be used.
        precision (int | str | torch.dtype | None):
            The precision of the index. If `None`, the default precision will be used.
        task (TaskType):
            The type of task for the index.
        **kwargs:
            Additional keyword arguments to be passed to the index instantiation.

    Returns:
        Dict[TaskType, BaseDocumentIndex]:
            A dictionary containing the loaded document index for each task type.

    Raises:
        ValueError: If the `index` parameter is not of type `BaseDocumentIndex`, `DictConfig`, or `Dict`.
        ValueError: If the `index` parameter is a `DictConfig` without a `_target_` key.
        ValueError: If the task type specified in the `index` parameter is not valid.
    """

    # index
    _index: Dict[TaskType, BaseDocumentIndex] = {
        TaskType.SPAN: None,
        TaskType.TRIPLET: None,
    }

    # check retriever type, it can be a BaseDocumentIndex, a DictConfig or a Dict
    if not isinstance(index, (BaseDocumentIndex, DictConfig, Dict, str)):
        raise ValueError(
            f"`index` must be a `BaseDocumentIndex`, a `DictConfig`, "
            f"a `Dict`, or a `str`, got `{type(index)}`."
        )
    # we need to check weather the DictConfig is a DictConfig for an instance of BaseDocumentIndex
    # or a primitive Dict
    if isinstance(index, str):
        # use_faiss = kwargs.get("use_faiss", False)
        index = {"name_or_path": index}
        # if use_faiss:
        #     index["_target_"] = "relik.retriever.indexers.faissindex.FaissDocumentIndex.from_pretrained"

    # if isinstance(index, DictConfig):
    #     # then it is probably a primitive Dict
    #     if "_target_" not in index:
    #         index = OmegaConf.to_container(index, resolve=True)
    #         # convert the key to TaskType
    #         try:
    #             index = {TaskType(k.lower()): v for k, v in index.items()}
    #         except ValueError as e:
    #             raise ValueError(
    #                 f"Please choose a valid task type (one of {list(TaskType)}) for each index."
    #             ) from e

    try:
        # convert the key to TaskType
        index = {TaskType(k): v for k, v in index.items()}
    except ValueError as e:
        index = {task: index}

    # instantiate each retriever
    if task in [TaskType.SPAN, TaskType.BOTH]:
        _index[TaskType.SPAN] = _instantiate_index(
            index[TaskType.SPAN],
            device,
            precision,
            **kwargs,
        )
    if task in [TaskType.TRIPLET, TaskType.BOTH]:
        _index[TaskType.TRIPLET] = _instantiate_index(
            index[TaskType.TRIPLET],
            device,
            precision,
            **kwargs,
        )
    # clean up None retrievers from the dictionary
    _index = {task_type: i for task_type, i in _index.items() if i is not None}
    return _index


def load_reader(
    reader: RelikReaderBase,
    device: str | None,
    precision: int | str | torch.dtype | None,
    compile: bool = False,
    **kwargs: Dict[str, Any],
) -> RelikReaderBase:
    """
    Load a reader model for inference.

    Args:
        reader (RelikReaderBase):
            The reader model to load.
        device (str | None):
            The device to move the reader model to.
        precision (int | str | torch.dtype | None):
            The precision to set for the reader model.
        compile (bool, optional):
            Whether to compile the reader model. Defaults to False.
        **kwargs (Dict[str, Any]):
            Additional keyword arguments to pass to the reader model.

    Returns:
        RelikReaderBase: The loaded reader model.
    """

    if not isinstance(reader, (RelikReaderBase, DictConfig, Dict, str)):
        raise ValueError(
            f"`reader` must be a `RelikReaderBase`, a `DictConfig`, "
            f"a `Dict`, or a `str`, got `{type(reader)}`."
        )

    if isinstance(reader, str):
        reader = {
            "_target_": "relik.reader.pytorch_modules.base.RelikReaderBase.from_pretrained",
            "model_name_or_dir": reader,
        }

    if not isinstance(reader, DictConfig):
        # then it is probably a primitive Dict
        # if "_target_" not in reader:
        #     reader = OmegaConf.to_container(reader, resolve=True)
        # reader = OmegaConf.to_container(reader, resolve=True)
        # if not isinstance(reader, DictConfig):
        reader = OmegaConf.create(reader)

    reader = (
        hydra.utils.instantiate(
            reader,
            device=device,
            precision=precision,
            **kwargs,
        )
        if isinstance(reader, DictConfig)
        else reader
    )
    reader.training = False
    reader.eval()
    if device is not None:
        logger.info(f"Moving reader to `{device}`.")
        reader.to(device)
    if precision is not None and reader.precision != PRECISION_MAP[precision]:
        logger.info(f"Setting precision of reader to `{PRECISION_MAP[precision]}`.")
        reader.to(PRECISION_MAP[precision])

    if compile:
        reader = torch.compile(reader)
    return reader



from typing import Any, Dict, Optional

import torch
from transformers import AutoModel, PreTrainedModel
from transformers.activations import ClippedGELUActivation, GELUActivation
from transformers.configuration_utils import PretrainedConfig
from transformers.modeling_utils import PoolerEndLogits

from .configuration_relik import RelikReaderConfig


class RelikReaderSample:
    def __init__(self, **kwargs):
        super().__setattr__("_d", {})
        self._d = kwargs

    def __getattribute__(self, item):
        return super(RelikReaderSample, self).__getattribute__(item)

    def __getattr__(self, item):
        if item.startswith("__") and item.endswith("__"):
            # this is likely some python library-specific variable (such as __deepcopy__ for copy)
            # better follow standard behavior here
            raise AttributeError(item)
        elif item in self._d:
            return self._d[item]
        else:
            return None

    def __setattr__(self, key, value):
        if key in self._d:
            self._d[key] = value
        else:
            super().__setattr__(key, value)
            self._d[key] = value


activation2functions = {
    "relu": torch.nn.ReLU(),
    "gelu": GELUActivation(),
    "gelu_10": ClippedGELUActivation(-10, 10),
}


class PoolerEndLogitsBi(PoolerEndLogits):
    def __init__(self, config: PretrainedConfig):
        super().__init__(config)
        self.dense_1 = torch.nn.Linear(config.hidden_size, 2)

    def forward(
        self,
        hidden_states: torch.FloatTensor,
        start_states: Optional[torch.FloatTensor] = None,
        start_positions: Optional[torch.LongTensor] = None,
        p_mask: Optional[torch.FloatTensor] = None,
    ) -> torch.FloatTensor:
        if p_mask is not None:
            p_mask = p_mask.unsqueeze(-1)
        logits = super().forward(
            hidden_states,
            start_states,
            start_positions,
            p_mask,
        )
        return logits


class RelikReaderSpanModel(PreTrainedModel):
    config_class = RelikReaderConfig

    def __init__(self, config: RelikReaderConfig, *args, **kwargs):
        super().__init__(config)
        # Transformer model declaration
        self.config = config
        self.transformer_model = (
            AutoModel.from_pretrained(self.config.transformer_model)
            if self.config.num_layers is None
            else AutoModel.from_pretrained(
                self.config.transformer_model, num_hidden_layers=self.config.num_layers
            )
        )
        self.transformer_model.resize_token_embeddings(
            self.transformer_model.config.vocab_size
            + self.config.additional_special_symbols
        )

        self.activation = self.config.activation
        self.linears_hidden_size = self.config.linears_hidden_size
        self.use_last_k_layers = self.config.use_last_k_layers

        # named entity detection layers
        self.ned_start_classifier = self._get_projection_layer(
            self.activation, last_hidden=2, layer_norm=False
        )
        if self.config.binary_end_logits:
            self.ned_end_classifier = PoolerEndLogitsBi(self.transformer_model.config)
        else:
            self.ned_end_classifier = PoolerEndLogits(self.transformer_model.config)

        # END entity disambiguation layer
        self.ed_start_projector = self._get_projection_layer(self.activation)
        self.ed_end_projector = self._get_projection_layer(self.activation)

        self.training = self.config.training

        # criterion
        self.criterion = torch.nn.CrossEntropyLoss()

    def _get_projection_layer(
        self,
        activation: str,
        last_hidden: Optional[int] = None,
        input_hidden=None,
        layer_norm: bool = True,
    ) -> torch.nn.Sequential:
        head_components = [
            torch.nn.Dropout(0.1),
            torch.nn.Linear(
                (
                    self.transformer_model.config.hidden_size * self.use_last_k_layers
                    if input_hidden is None
                    else input_hidden
                ),
                self.linears_hidden_size,
            ),
            activation2functions[activation],
            torch.nn.Dropout(0.1),
            torch.nn.Linear(
                self.linears_hidden_size,
                self.linears_hidden_size if last_hidden is None else last_hidden,
            ),
        ]

        if layer_norm:
            head_components.append(
                torch.nn.LayerNorm(
                    self.linears_hidden_size if last_hidden is None else last_hidden,
                    self.transformer_model.config.layer_norm_eps,
                )
            )

        return torch.nn.Sequential(*head_components)

    def _mask_logits(self, logits: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        mask = mask.unsqueeze(-1)
        if next(self.parameters()).dtype == torch.float16:
            logits = logits * (1 - mask) - 65500 * mask
        else:
            logits = logits * (1 - mask) - 1e30 * mask
        return logits

    def _get_model_features(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        token_type_ids: Optional[torch.Tensor],
    ):
        model_input = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "output_hidden_states": self.use_last_k_layers > 1,
        }

        if token_type_ids is not None:
            model_input["token_type_ids"] = token_type_ids

        model_output = self.transformer_model(**model_input)

        if self.use_last_k_layers > 1:
            model_features = torch.cat(
                model_output[1][-self.use_last_k_layers :], dim=-1
            )
        else:
            model_features = model_output[0]

        return model_features

    def compute_ned_end_logits(
        self,
        start_predictions,
        start_labels,
        model_features,
        prediction_mask,
        batch_size,
    ) -> Optional[torch.Tensor]:
        # todo: maybe when constraining on the spans,
        #  we should not use a prediction_mask for the end tokens.
        #  at least we should not during training imo
        start_positions = start_labels if self.training else start_predictions
        start_positions_indices = (
            torch.arange(start_positions.size(1), device=start_positions.device)
            .unsqueeze(0)
            .expand(batch_size, -1)[start_positions > 0]
        ).to(start_positions.device)

        if len(start_positions_indices) > 0:
            expanded_features = model_features.repeat_interleave(
                torch.sum(start_positions > 0, dim=-1), dim=0
            )
            expanded_prediction_mask = prediction_mask.repeat_interleave(
                torch.sum(start_positions > 0, dim=-1), dim=0
            )
            end_logits = self.ned_end_classifier(
                hidden_states=expanded_features,
                start_positions=start_positions_indices,
                p_mask=expanded_prediction_mask,
            )

            return end_logits

        return None

    def compute_classification_logits(
        self,
        model_features_start,
        model_features_end,
        special_symbols_features,
    ) -> torch.Tensor:
        model_start_features = self.ed_start_projector(model_features_start)
        model_end_features = self.ed_end_projector(model_features_end)
        model_start_features_symbols = self.ed_start_projector(special_symbols_features)
        model_end_features_symbols = self.ed_end_projector(special_symbols_features)

        model_ed_features = torch.cat(
            [model_start_features, model_end_features], dim=-1
        )
        special_symbols_representation = torch.cat(
            [model_start_features_symbols, model_end_features_symbols], dim=-1
        )

        logits = torch.bmm(
            model_ed_features,
            torch.permute(special_symbols_representation, (0, 2, 1)),
        )

        logits = self._mask_logits(logits, (model_features_start == -100).all(2).long())
        return logits

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        token_type_ids: Optional[torch.Tensor] = None,
        prediction_mask: Optional[torch.Tensor] = None,
        special_symbols_mask: Optional[torch.Tensor] = None,
        start_labels: Optional[torch.Tensor] = None,
        end_labels: Optional[torch.Tensor] = None,
        use_predefined_spans: bool = False,
        *args,
        **kwargs,
    ) -> Dict[str, Any]:
        batch_size, seq_len = input_ids.shape

        model_features = self._get_model_features(
            input_ids, attention_mask, token_type_ids
        )

        ned_start_labels = None

        # named entity detection if required
        if use_predefined_spans:  # no need to compute spans
            ned_start_logits, ned_start_probabilities, ned_start_predictions = (
                None,
                None,
                (
                    torch.clone(start_labels)
                    if start_labels is not None
                    else torch.zeros_like(input_ids)
                ),
            )
            ned_end_logits, ned_end_probabilities, ned_end_predictions = (
                None,
                None,
                (
                    torch.clone(end_labels)
                    if end_labels is not None
                    else torch.zeros_like(input_ids)
                ),
            )
            ned_start_predictions[ned_start_predictions > 0] = 1
            ned_end_predictions[end_labels > 0] = 1
            ned_end_predictions = ned_end_predictions[~(end_labels == -100).all(2)]

        else:  # compute spans
            # start boundary prediction
            ned_start_logits = self.ned_start_classifier(model_features)
            ned_start_logits = self._mask_logits(ned_start_logits, prediction_mask)
            ned_start_probabilities = torch.softmax(ned_start_logits, dim=-1)
            ned_start_predictions = ned_start_probabilities.argmax(dim=-1)

            # end boundary prediction
            ned_start_labels = (
                torch.zeros_like(start_labels) if start_labels is not None else None
            )

            if ned_start_labels is not None:
                ned_start_labels[start_labels == -100] = -100
                ned_start_labels[start_labels > 0] = 1

            ned_end_logits = self.compute_ned_end_logits(
                ned_start_predictions,
                ned_start_labels,
                model_features,
                prediction_mask,
                batch_size,
            )

            if ned_end_logits is not None:
                ned_end_probabilities = torch.softmax(ned_end_logits, dim=-1)
                if not self.config.binary_end_logits:
                    ned_end_predictions = torch.argmax(
                        ned_end_probabilities, dim=-1, keepdim=True
                    )
                    ned_end_predictions = torch.zeros_like(
                        ned_end_probabilities
                    ).scatter_(1, ned_end_predictions, 1)
                else:
                    ned_end_predictions = torch.argmax(ned_end_probabilities, dim=-1)
            else:
                ned_end_logits, ned_end_probabilities = None, None
                ned_end_predictions = ned_start_predictions.new_zeros(
                    batch_size, seq_len
                )

            if not self.training:
                # if len(ned_end_predictions.shape) < 2:
                #     print(ned_end_predictions)
                end_preds_count = ned_end_predictions.sum(1)
                # If there are no end predictions for a start prediction, remove the start prediction
                if (end_preds_count == 0).any() and (ned_start_predictions > 0).any():
                    ned_start_predictions[ned_start_predictions == 1] = (
                        end_preds_count != 0
                    ).long()
                    ned_end_predictions = ned_end_predictions[end_preds_count != 0]

        if end_labels is not None:
            end_labels = end_labels[~(end_labels == -100).all(2)]

        start_position, end_position = (
            (start_labels, end_labels)
            if self.training
            else (ned_start_predictions, ned_end_predictions)
        )
        start_counts = (start_position > 0).sum(1)
        if (start_counts > 0).any():
            ned_end_predictions = ned_end_predictions.split(start_counts.tolist())
        # Entity disambiguation
        if (end_position > 0).sum() > 0:
            ends_count = (end_position > 0).sum(1)
            model_entity_start = torch.repeat_interleave(
                model_features[start_position > 0], ends_count, dim=0
            )
            model_entity_end = torch.repeat_interleave(
                model_features, start_counts, dim=0
            )[end_position > 0]
            ents_count = torch.nn.utils.rnn.pad_sequence(
                torch.split(ends_count, start_counts.tolist()),
                batch_first=True,
                padding_value=0,
            ).sum(1)

            model_entity_start = torch.nn.utils.rnn.pad_sequence(
                torch.split(model_entity_start, ents_count.tolist()),
                batch_first=True,
                padding_value=-100,
            )

            model_entity_end = torch.nn.utils.rnn.pad_sequence(
                torch.split(model_entity_end, ents_count.tolist()),
                batch_first=True,
                padding_value=-100,
            )

            ed_logits = self.compute_classification_logits(
                model_entity_start,
                model_entity_end,
                model_features[special_symbols_mask].view(
                    batch_size, -1, model_features.shape[-1]
                ),
            )
            ed_probabilities = torch.softmax(ed_logits, dim=-1)
            ed_predictions = torch.argmax(ed_probabilities, dim=-1)
        else:
            ed_logits, ed_probabilities, ed_predictions = (
                None,
                ned_start_predictions.new_zeros(batch_size, seq_len),
                ned_start_predictions.new_zeros(batch_size),
            )
        # output build
        output_dict = dict(
            batch_size=batch_size,
            ned_start_logits=ned_start_logits,
            ned_start_probabilities=ned_start_probabilities,
            ned_start_predictions=ned_start_predictions,
            ned_end_logits=ned_end_logits,
            ned_end_probabilities=ned_end_probabilities,
            ned_end_predictions=ned_end_predictions,
            ed_logits=ed_logits,
            ed_probabilities=ed_probabilities,
            ed_predictions=ed_predictions,
        )

        # compute loss if labels
        if start_labels is not None and end_labels is not None and self.training:
            # named entity detection loss

            # start
            if ned_start_logits is not None:
                ned_start_loss = self.criterion(
                    ned_start_logits.view(-1, ned_start_logits.shape[-1]),
                    ned_start_labels.view(-1),
                )
            else:
                ned_start_loss = 0

            # end
            # use ents_count to assign the labels to the correct positions i.e. using end_labels -> [[0,0,4,0], [0,0,0,2]] -> [4,2] (this is just an element, for batch we need to mask it with ents_count), ie -> [[4,2,-100,-100], [3,1,2,-100], [1,3,2,5]]

            if ned_end_logits is not None:
                ed_labels = end_labels.clone()
                ed_labels = torch.nn.utils.rnn.pad_sequence(
                    torch.split(ed_labels[ed_labels > 0], ents_count.tolist()),
                    batch_first=True,
                    padding_value=-100,
                )
                end_labels[end_labels > 0] = 1
                if not self.config.binary_end_logits:
                    # transform label to position in the sequence
                    end_labels = end_labels.argmax(dim=-1)
                    ned_end_loss = self.criterion(
                        ned_end_logits.view(-1, ned_end_logits.shape[-1]),
                        end_labels.view(-1),
                    )
                else:
                    ned_end_loss = self.criterion(
                        ned_end_logits.reshape(-1, ned_end_logits.shape[-1]),
                        end_labels.reshape(-1).long(),
                    )

                # entity disambiguation loss
                ed_loss = self.criterion(
                    ed_logits.view(-1, ed_logits.shape[-1]),
                    ed_labels.view(-1).long(),
                )

            else:
                ned_end_loss = 0
                ed_loss = 0

            output_dict["ned_start_loss"] = ned_start_loss
            output_dict["ned_end_loss"] = ned_end_loss
            output_dict["ed_loss"] = ed_loss

            output_dict["loss"] = ned_start_loss + ned_end_loss + ed_loss

        return output_dict


class RelikReaderREModel(PreTrainedModel):
    config_class = RelikReaderConfig

    def __init__(self, config, *args, **kwargs):
        super().__init__(config)
        # Transformer model declaration
        # self.transformer_model_name = transformer_model
        self.config = config
        self.transformer_model = (
            AutoModel.from_pretrained(config.transformer_model)
            if config.num_layers is None
            else AutoModel.from_pretrained(
                config.transformer_model, num_hidden_layers=config.num_layers
            )
        )
        self.transformer_model.resize_token_embeddings(
            self.transformer_model.config.vocab_size
            + config.additional_special_symbols
            + config.additional_special_symbols_types,
        )

        # named entity detection layers
        self.ned_start_classifier = self._get_projection_layer(
            config.activation, last_hidden=2, layer_norm=False
        )

        self.ned_end_classifier = PoolerEndLogitsBi(self.transformer_model.config)

        self.relation_disambiguation_loss = (
            config.relation_disambiguation_loss
            if hasattr(config, "relation_disambiguation_loss")
            else False
        )

        if self.config.entity_type_loss and self.config.add_entity_embedding:
            input_hidden_ents = 3
        else:
            input_hidden_ents = 2

        self.re_projector = self._get_projection_layer(
            config.activation,
            input_hidden=input_hidden_ents * self.transformer_model.config.hidden_size,
            hidden=input_hidden_ents * self.config.linears_hidden_size,
            last_hidden=2 * self.config.linears_hidden_size,
        )

        self.re_relation_projector = self._get_projection_layer(
            config.activation,
            input_hidden=self.transformer_model.config.hidden_size,
        )

        if self.config.entity_type_loss or self.relation_disambiguation_loss:
            self.re_entities_projector = self._get_projection_layer(
                config.activation,
                input_hidden=2 * self.transformer_model.config.hidden_size,
            )
            self.re_definition_projector = self._get_projection_layer(
                config.activation,
            )

        self.re_classifier = self._get_projection_layer(
            config.activation,
            input_hidden=config.linears_hidden_size,
            last_hidden=2,
            layer_norm=False,
        )

        self.training = config.training

        # criterion
        self.criterion = torch.nn.CrossEntropyLoss()
        self.criterion_type = torch.nn.BCEWithLogitsLoss()

    def _get_projection_layer(
        self,
        activation: str,
        last_hidden: Optional[int] = None,
        hidden: Optional[int] = None,
        input_hidden=None,
        layer_norm: bool = True,
    ) -> torch.nn.Sequential:
        head_components = [
            torch.nn.Dropout(0.1),
            torch.nn.Linear(
                (
                    self.transformer_model.config.hidden_size
                    * self.config.use_last_k_layers
                    if input_hidden is None
                    else input_hidden
                ),
                self.config.linears_hidden_size if hidden is None else hidden,
            ),
            activation2functions[activation],
            torch.nn.Dropout(0.1),
            torch.nn.Linear(
                self.config.linears_hidden_size if hidden is None else hidden,
                self.config.linears_hidden_size if last_hidden is None else last_hidden,
            ),
        ]

        if layer_norm:
            head_components.append(
                torch.nn.LayerNorm(
                    (
                        self.config.linears_hidden_size
                        if last_hidden is None
                        else last_hidden
                    ),
                    self.transformer_model.config.layer_norm_eps,
                )
            )

        return torch.nn.Sequential(*head_components)

    def _mask_logits(self, logits: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        mask = mask.unsqueeze(-1)
        if next(self.parameters()).dtype == torch.float16:
            logits = logits * (1 - mask) - 65500 * mask
        else:
            logits = logits * (1 - mask) - 1e30 * mask
        return logits

    def _get_model_features(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        token_type_ids: Optional[torch.Tensor],
    ):
        model_input = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "output_hidden_states": self.config.use_last_k_layers > 1,
        }

        if token_type_ids is not None:
            model_input["token_type_ids"] = token_type_ids

        model_output = self.transformer_model(**model_input)

        if self.config.use_last_k_layers > 1:
            model_features = torch.cat(
                model_output[1][-self.config.use_last_k_layers :], dim=-1
            )
        else:
            model_features = model_output[0]

        return model_features

    def compute_ned_end_logits(
        self,
        start_predictions,
        start_labels,
        model_features,
        prediction_mask,
        batch_size,
        mask_preceding: bool = False,
    ) -> Optional[torch.Tensor]:
        # todo: maybe when constraining on the spans,
        #  we should not use a prediction_mask for the end tokens.
        #  at least we should not during training imo
        start_positions = start_labels if self.training else start_predictions
        start_positions_indices = (
            torch.arange(start_positions.size(1), device=start_positions.device)
            .unsqueeze(0)
            .expand(batch_size, -1)[start_positions > 0]
        ).to(start_positions.device)

        if len(start_positions_indices) > 0:
            expanded_features = model_features.repeat_interleave(
                torch.sum(start_positions > 0, dim=-1), dim=0
            )
            expanded_prediction_mask = prediction_mask.repeat_interleave(
                torch.sum(start_positions > 0, dim=-1), dim=0
            )
            if mask_preceding:
                expanded_prediction_mask[
                    torch.arange(
                        expanded_prediction_mask.shape[1],
                        device=expanded_prediction_mask.device,
                    )
                    < start_positions_indices.unsqueeze(1)
                ] = 1
            end_logits = self.ned_end_classifier(
                hidden_states=expanded_features,
                start_positions=start_positions_indices,
                p_mask=expanded_prediction_mask,
            )

            return end_logits

        return None

    def compute_relation_logits(
        self,
        model_entity_features,
        special_symbols_features,
    ) -> torch.Tensor:
        model_subject_object_features = self.re_projector(model_entity_features)
        model_subject_features = model_subject_object_features[
            :, :, : model_subject_object_features.shape[-1] // 2
        ]
        model_object_features = model_subject_object_features[
            :, :, model_subject_object_features.shape[-1] // 2 :
        ]
        special_symbols_start_representation = self.re_relation_projector(
            special_symbols_features
        )
        re_logits = torch.einsum(
            "bse,bde,bfe->bsdfe",
            model_subject_features,
            model_object_features,
            special_symbols_start_representation,
        )
        re_logits = self.re_classifier(re_logits)

        return re_logits

    def compute_entity_logits(
        self,
        model_entity_features,
        special_symbols_features,
    ) -> torch.Tensor:
        model_ed_features = self.re_entities_projector(model_entity_features)
        special_symbols_ed_representation = self.re_definition_projector(
            special_symbols_features
        )

        logits = torch.bmm(
            model_ed_features,
            torch.permute(special_symbols_ed_representation, (0, 2, 1)),
        )
        logits = self._mask_logits(
            logits, (model_entity_features == -100).all(2).long()
        )
        return logits

    def compute_loss(self, logits, labels, mask=None):
        logits = logits.reshape(-1, logits.shape[-1])
        labels = labels.reshape(-1).long()
        if mask is not None:
            return self.criterion(logits[mask], labels[mask])
        return self.criterion(logits, labels)

    def compute_ned_type_loss(
        self,
        disambiguation_labels,
        re_ned_entities_logits,
        ned_type_logits,
        re_entities_logits,
        entity_types,
        mask,
    ):
        if self.config.entity_type_loss and self.relation_disambiguation_loss:
            return self.criterion_type(
                re_ned_entities_logits[disambiguation_labels != -100],
                disambiguation_labels[disambiguation_labels != -100],
            )
        if self.config.entity_type_loss:
            return self.criterion_type(
                ned_type_logits[mask],
                disambiguation_labels[:, :, :entity_types][mask],
            )

        if self.relation_disambiguation_loss:
            return self.criterion_type(
                re_entities_logits[disambiguation_labels != -100],
                disambiguation_labels[disambiguation_labels != -100],
            )
        return 0

    def compute_relation_loss(self, relation_labels, re_logits):
        return self.compute_loss(
            re_logits, relation_labels, relation_labels.view(-1) != -100
        )

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        token_type_ids: torch.Tensor,
        prediction_mask: Optional[torch.Tensor] = None,
        special_symbols_mask: Optional[torch.Tensor] = None,
        special_symbols_mask_entities: Optional[torch.Tensor] = None,
        start_labels: Optional[torch.Tensor] = None,
        end_labels: Optional[torch.Tensor] = None,
        disambiguation_labels: Optional[torch.Tensor] = None,
        relation_labels: Optional[torch.Tensor] = None,
        relation_threshold: float = None,
        is_validation: bool = False,
        is_prediction: bool = False,
        use_predefined_spans: bool = False,
        *args,
        **kwargs,
    ) -> Dict[str, Any]:
        relation_threshold = (
            self.config.threshold if relation_threshold is None else relation_threshold
        )

        batch_size = input_ids.shape[0]

        model_features = self._get_model_features(
            input_ids, attention_mask, token_type_ids
        )

        # named entity detection
        if use_predefined_spans:
            ned_start_logits, ned_start_probabilities, ned_start_predictions = (
                None,
                None,
                torch.zeros_like(start_labels),
            )
            ned_end_logits, ned_end_probabilities, ned_end_predictions = (
                None,
                None,
                torch.zeros_like(end_labels),
            )

            ned_start_predictions[start_labels > 0] = 1
            ned_end_predictions[end_labels > 0] = 1
            ned_end_predictions = ned_end_predictions[~(end_labels == -100).all(2)]
            ned_start_labels = start_labels
            ned_start_labels[start_labels > 0] = 1
        else:
            # start boundary prediction
            ned_start_logits = self.ned_start_classifier(model_features)
            if is_validation or is_prediction:
                ned_start_logits = self._mask_logits(
                    ned_start_logits, prediction_mask
                )  # why?
            ned_start_probabilities = torch.softmax(ned_start_logits, dim=-1)
            ned_start_predictions = ned_start_probabilities.argmax(dim=-1)

            # end boundary prediction
            ned_start_labels = (
                torch.zeros_like(start_labels) if start_labels is not None else None
            )

            # start_labels contain entity id at their position, we just need 1 for start of entity
            if ned_start_labels is not None:
                ned_start_labels[start_labels == -100] = -100
                ned_start_labels[start_labels > 0] = 1

            # compute end logits only if there are any start predictions.
            # For each start prediction, n end predictions are made
            ned_end_logits = self.compute_ned_end_logits(
                ned_start_predictions,
                ned_start_labels,
                model_features,
                prediction_mask,
                batch_size,
                True,
            )

            if ned_end_logits is not None:
                # For each start prediction, n end predictions are made based on
                # binary classification ie. argmax at each position.
                ned_end_probabilities = torch.softmax(ned_end_logits, dim=-1)
                ned_end_predictions = ned_end_probabilities.argmax(dim=-1)
            else:
                ned_end_logits, ned_end_probabilities = None, None
                ned_end_predictions = torch.zeros_like(ned_start_predictions)

            if is_prediction or is_validation:
                end_preds_count = ned_end_predictions.sum(1)
                # If there are no end predictions for a start prediction, remove the start prediction
                if (end_preds_count == 0).any() and (ned_start_predictions > 0).any():
                    ned_start_predictions[ned_start_predictions == 1] = (
                        end_preds_count != 0
                    ).long()
                    ned_end_predictions = ned_end_predictions[end_preds_count != 0]

        if end_labels is not None:
            end_labels = end_labels[~(end_labels == -100).all(2)]

        start_position, end_position = (
            (start_labels, end_labels)
            if (not is_prediction and not is_validation)
            else (ned_start_predictions, ned_end_predictions)
        )

        start_counts = (start_position > 0).sum(1)
        if (start_counts > 0).any():
            ned_end_predictions = ned_end_predictions.split(start_counts.tolist())
        else:
            ned_end_predictions = [torch.empty(0, input_ids.shape[1], dtype=torch.int64) for _ in range(batch_size)]
        # limit to 30 predictions per document using start_counts, by setting all po after sum is 30 to 0
        # if is_validation or is_prediction:
        #     ned_start_predictions[ned_start_predictions == 1] = start_counts
        # We can only predict relations if we have start and end predictions
        if (end_position > 0).sum() > 0:
            ends_count = (end_position > 0).sum(1)
            model_subject_features = torch.cat(
                [
                    torch.repeat_interleave(
                        model_features[start_position > 0], ends_count, dim=0
                    ),  # start position features
                    torch.repeat_interleave(model_features, start_counts, dim=0)[
                        end_position > 0
                    ],  # end position features
                ],
                dim=-1,
            )
            ents_count = torch.nn.utils.rnn.pad_sequence(
                torch.split(ends_count, start_counts.tolist()),
                batch_first=True,
                padding_value=0,
            ).sum(1)
            model_subject_features = torch.nn.utils.rnn.pad_sequence(
                torch.split(model_subject_features, ents_count.tolist()),
                batch_first=True,
                padding_value=-100,
            )

            # if is_validation or is_prediction:
            #     model_subject_features = model_subject_features[:, :30, :]

            # entity disambiguation. Here relation_disambiguation_loss would only be useful to
            # reduce the number of candidate relations for the next step, but currently unused.
            if self.config.entity_type_loss or self.relation_disambiguation_loss:
                (re_ned_entities_logits) = self.compute_entity_logits(
                    model_subject_features,
                    model_features[
                        special_symbols_mask | special_symbols_mask_entities
                    ].view(batch_size, -1, model_features.shape[-1]),
                )
                entity_types = torch.sum(special_symbols_mask_entities, dim=1)[0].item()
                ned_type_logits = re_ned_entities_logits[:, :, :entity_types]
                re_entities_logits = re_ned_entities_logits[:, :, entity_types:]

                if self.config.entity_type_loss:
                    ned_type_probabilities = torch.sigmoid(ned_type_logits)
                    ned_type_predictions = ned_type_probabilities.argmax(dim=-1)

                    if self.config.add_entity_embedding:
                        special_symbols_representation = model_features[
                            special_symbols_mask_entities
                        ].view(batch_size, entity_types, -1)

                        entities_representation = torch.einsum(
                            "bsp,bpe->bse",
                            ned_type_probabilities,
                            special_symbols_representation,
                        )
                        model_subject_features = torch.cat(
                            [model_subject_features, entities_representation], dim=-1
                        )
                re_entities_probabilities = torch.sigmoid(re_entities_logits)
                re_entities_predictions = re_entities_probabilities.round()
            else:
                (
                    ned_type_logits,
                    ned_type_probabilities,
                    re_entities_logits,
                    re_entities_probabilities,
                ) = (None, None, None, None)
                ned_type_predictions, re_entities_predictions = (
                    torch.zeros([batch_size, 1], dtype=torch.long).to(input_ids.device),
                    torch.zeros([batch_size, 1], dtype=torch.long).to(input_ids.device),
                )

            # Compute relation logits
            re_logits = self.compute_relation_logits(
                model_subject_features,
                model_features[special_symbols_mask].view(
                    batch_size, -1, model_features.shape[-1]
                ),
            )

            re_probabilities = torch.softmax(re_logits, dim=-1)
            # we set a thresshold instead of argmax in cause it needs to be tweaked
            re_predictions = re_probabilities[:, :, :, :, 1] > relation_threshold
            re_probabilities = re_probabilities[:, :, :, :, 1]
        else:
            (
                ned_type_logits,
                ned_type_probabilities,
                re_entities_logits,
                re_entities_probabilities,
            ) = (None, None, None, None)
            ned_type_predictions, re_entities_predictions = (
                torch.zeros([batch_size, 1], dtype=torch.long).to(input_ids.device),
                torch.zeros([batch_size, 1], dtype=torch.long).to(input_ids.device),
            )
            re_logits, re_probabilities, re_predictions = (
                torch.zeros(
                    [batch_size, 1, 1, special_symbols_mask.sum(1)[0]], dtype=torch.long
                ).to(input_ids.device),
                torch.zeros(
                    [batch_size, 1, 1, special_symbols_mask.sum(1)[0]], dtype=torch.long
                ).to(input_ids.device),
                torch.zeros(
                    [batch_size, 1, 1, special_symbols_mask.sum(1)[0]], dtype=torch.long
                ).to(input_ids.device),
            )

        # output build
        output_dict = dict(
            batch_size=batch_size,
            ned_start_logits=ned_start_logits,
            ned_start_probabilities=ned_start_probabilities,
            ned_start_predictions=ned_start_predictions,
            ned_end_logits=ned_end_logits,
            ned_end_probabilities=ned_end_probabilities,
            ned_end_predictions=ned_end_predictions,
            ned_type_logits=ned_type_logits,
            ned_type_probabilities=ned_type_probabilities,
            ned_type_predictions=ned_type_predictions,
            re_entities_logits=re_entities_logits,
            re_entities_probabilities=re_entities_probabilities,
            re_entities_predictions=re_entities_predictions,
            re_logits=re_logits,
            re_probabilities=re_probabilities,
            re_predictions=re_predictions,
        )

        if (
            start_labels is not None
            and end_labels is not None
            and relation_labels is not None
            and is_prediction is False
        ):
            ned_start_loss = self.compute_loss(ned_start_logits, ned_start_labels)
            end_labels[end_labels > 0] = 1
            ned_end_loss = self.compute_loss(ned_end_logits, end_labels)
            if self.config.entity_type_loss or self.relation_disambiguation_loss:
                ned_type_loss = self.compute_ned_type_loss(
                    disambiguation_labels,
                    re_ned_entities_logits,
                    ned_type_logits,
                    re_entities_logits,
                    entity_types,
                    (model_subject_features != -100).all(2),
                )
            relation_loss = self.compute_relation_loss(relation_labels, re_logits)
            # compute loss. We can skip the relation loss if we are in the first epochs (optional)
            if self.config.entity_type_loss or self.relation_disambiguation_loss:
                output_dict["loss"] = (
                    ned_start_loss + ned_end_loss + relation_loss + ned_type_loss
                ) / 4
                output_dict["ned_type_loss"] = ned_type_loss
            else:
                output_dict["loss"] = ((1 / 20) * (ned_start_loss + ned_end_loss)) + (
                    (9 / 10) * relation_loss
                )
            output_dict["ned_start_loss"] = ned_start_loss
            output_dict["ned_end_loss"] = ned_end_loss
            output_dict["re_loss"] = relation_loss

        return output_dict



import csv
import json
import sys
import time
from pathlib import Path
from typing import Optional

import hydra
import torch
import tqdm
import typer

from relik.cli.utils import resolve_config
from relik.common.log import get_logger, print_relik_text_art
from relik.common.utils import get_callable_from_string
from relik.retriever import GoldenRetriever
from relik.retriever.common.model_inputs import ModelInputs
from relik.retriever.data.base.datasets import BaseDataset
from relik.retriever.indexers.document import DocumentStore
from relik.retriever.trainer.train import train_hydra as retriever_train

logger = get_logger(__name__)

app = typer.Typer(no_args_is_help=True, pretty_exceptions_show_locals=False)


@app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
def train():
    """
    Trains the retriever model.

    This function prints the Relik text art, resolves the configuration file path,
    and then calls the `_retriever_train` function to train the retriever model.

    Args:
        None

    Returns:
        None
    """
    print_relik_text_art()
    config_dir, config_name, overrides = resolve_config("retriever")

    @hydra.main(
        config_path=str(config_dir),
        config_name=str(config_name),
        version_base="1.3",
    )
    def _retriever_train(conf):
        retriever_train(conf)

    # clean sys.argv for hydra
    sys.argv = sys.argv[:1]
    # add the overrides to sys.argv
    sys.argv.extend(overrides)

    _retriever_train()


@torch.no_grad()
@app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
def create_index(
    question_encoder_name_or_path: str,
    document_path: str,
    output_folder: str,
    document_file_type: str = "jsonl",
    passage_encoder_name_or_path: Optional[str] = None,
    indexer_class: str = "relik.retriever.indexers.inmemory.InMemoryDocumentIndex",
    batch_size: int = 512,
    num_workers: int = 4,
    passage_max_length: int = 64,
    device: str = "cuda",
    index_device: str = "cpu",
    precision: str = "fp32",
    push_to_hub: bool = False,
    repo_id: Optional[str] = None,
):
    """
    Builds an index for document retrieval.

    Args:
        question_encoder_name_or_path (str):
            The name or path of the question encoder model.
        document_path (str):
            The path to the document file.
        output_folder (str):
            The folder where the index will be saved.
        document_file_type (str, optional):
            The type of the document file. Defaults to "jsonl".
        passage_encoder_name_or_path (str, optional):
            The name or path of the passage encoder model. Defaults to None.
        indexer_class (str, optional):
            The class of the document indexer. Defaults to "relik.retriever.indexers.inmemory.InMemoryDocumentIndex".
        batch_size (int, optional):
            The batch size for indexing. Defaults to 512.
        num_workers (int, optional):
            The number of workers for indexing. Defaults to 4.
        passage_max_length (int, optional):
            The maximum length of a passage. Defaults to 64.
        device (str, optional):
            The device to use for indexing. Defaults to "cuda".
        index_device (str, optional):
            The device to use for indexing the document index. Defaults to "cpu".
        precision (str, optional):
            The precision for indexing. Defaults to "fp32".
        push_to_hub (bool, optional):
             Whether to push the index to the Hugging Face Model Hub. Defaults to False.
        repo_id (str, optional):
            The ID of the repository in the Hugging Face Model Hub. Required if push_to_hub is True.

    Raises:
        ValueError: If `repo_id` is not provided when `push_to_hub=True`.

    Returns:
        None
    """

    if push_to_hub:
        if not repo_id:
            raise ValueError("`repo_id` must be provided when `push_to_hub=True`")

    logger.info("Loading documents")
    if document_file_type == "jsonl":
        documents = DocumentStore.from_file(document_path)
    elif document_file_type == "csv":
        documents = DocumentStore.from_tsv(
            document_path, delimiter=",", quoting=csv.QUOTE_NONE, ingore_case=True
        )
    elif document_file_type == "tsv":
        documents = DocumentStore.from_tsv(
            document_path, delimiter="\t", quoting=csv.QUOTE_NONE, ingore_case=True
        )
    else:
        raise ValueError(
            f"Unknown document file type: {document_file_type}, must be one of jsonl, csv, tsv"
        )

    logger.info("Loading document index")
    logger.info(f"Loaded {len(documents)} documents")
    indexer = get_callable_from_string(indexer_class)(
        documents, device=index_device, precision=precision
    )

    retriever = GoldenRetriever(
        question_encoder=question_encoder_name_or_path,
        passage_encoder=passage_encoder_name_or_path,
        document_index=indexer,
        device=device,
        precision=precision,
    )
    retriever.eval()

    retriever.index(
        batch_size=batch_size,
        num_workers=num_workers,
        max_length=passage_max_length,
        force_reindex=True,
        precision=precision,
    )

    output_folder = Path(output_folder)
    output_folder.mkdir(exist_ok=True, parents=True)
    retriever.document_index.save_pretrained(
        output_folder, push_to_hub=push_to_hub, model_id=repo_id
    )


@torch.no_grad()
@app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
def add_candidates(
    question_encoder_name_or_path: str,
    document_name_or_path: str,
    input_path: str,
    output_path: str,
    passage_encoder_name_or_path: Optional[str] = None,
    relations: bool = False,
    top_k: int = 100,
    batch_size: int = 128,
    num_workers: int = 4,
    device: str = "cuda",
    index_device: str = "cpu",
    precision: str = "fp32",
    use_doc_topics: bool = False,
    log_recall: bool = False,
):
    """
    Adds candidates to the input samples based on retrieval from a document index.

    Args:
        question_encoder_name_or_path (str):
            The name or path of the question encoder model.
        document_name_or_path (str):
            The name or path of the document index.
        input_path (str):
            The path to the input file containing samples.
        output_path (str):
            The path to the output file where the samples with candidates will be saved.
        passage_encoder_name_or_path (Optional[str]):
            The name or path of the passage encoder model. Defaults to None.
        relations (bool):
            Whether to add the candidates as relations. Defaults to False.
        top_k (int):
            The number of candidates to retrieve for each sample. Defaults to 100.
        batch_size (int):
            The batch size for retrieval. Defaults to 128.
        num_workers (int):
            The number of worker processes for data loading. Defaults to 4.
        device (str):
            The device to use for retrieval. Defaults to "cuda".
        index_device (str):
            The device to use for the document index. Defaults to "cpu".
        precision (str):
            The precision to use for retrieval. Defaults to "fp32".
        use_doc_topics (bool):
            Whether to use document topics for retrieval. Defaults to False.
        log_recall (bool):
            Whether to log the recall of the retrieval. Defaults to False.

    Raises:
        ValueError: If the dataset does not contain topics but `use_doc_topics` is set to True.

    Returns:
        None
    """
    retriever = GoldenRetriever(
        question_encoder=question_encoder_name_or_path,
        passage_encoder=passage_encoder_name_or_path,
        document_index=document_name_or_path,
        device=device,
        index_device=index_device,
        index_precision=precision,
    )
    retriever.eval()

    logger.info(f"Loading from {input_path}")
    with open(input_path) as f:
        samples = [json.loads(line) for line in f.readlines()]

    if use_doc_topics and "doc_topic" not in samples[0]:
        raise ValueError(
            "Dataset does not contain topics, but --use-doc-topics was passed"
        )
    use_doc_topics = use_doc_topics and "doc_topic" in samples[0]

    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    correct, total = 0, 0

    with open(output_path, "w") as f_out:
        # get tokenizer
        tokenizer = retriever.question_tokenizer

        def collate_fn(batch):
            return ModelInputs(
                tokenizer(
                    [b["text"] for b in batch],
                    text_pair=(
                        [b["doc_topic"] for b in batch] if use_doc_topics else None
                    ),
                    padding=True,
                    return_tensors="pt",
                    truncation=True,
                )
            )

        logger.info(
            f"Creating dataloader with batch size {batch_size} and {num_workers} workers"
        )
        dataloader = torch.utils.data.DataLoader(
            BaseDataset(name="passage", data=samples),
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            pin_memory=False,
            collate_fn=collate_fn,
        )

        # we also dump the candidates to a file after a while
        retrieved_accumulator = []
        with torch.inference_mode():
            num_completed_docs = 0

            start = time.time()
            for documents_batch in tqdm.tqdm(dataloader):
                retrieve_kwargs = {
                    **documents_batch,
                    "k": top_k,
                    "precision": precision,
                }
                batch_out = retriever.retrieve(**retrieve_kwargs)
                retrieved_accumulator.extend(batch_out)

                if len(retrieved_accumulator) % 1_000 == 0:
                    output_data = []
                    # get the correct document from the original dataset
                    # the dataloader is not shuffled, so we can just count the number of
                    # documents we have seen so far
                    for sample, retrieved in zip(
                        samples[
                            num_completed_docs : num_completed_docs
                            + len(retrieved_accumulator)
                        ],
                        retrieved_accumulator,
                    ):
                        candidate_titles = [
                            c.document.text
                            for c in retrieved  # TODO: add metadata if needed
                        ]
                        # TODO: compatibility shit
                        if relations:
                            sample["triplet_candidates"] = candidate_titles
                            sample["triplet_candidates_scores"] = [
                                c.score for c in retrieved
                            ]
                            if log_recall:
                                for triplet in sample["window_triplet_labels"]:
                                    relation = triplet["relation"]
                                    if relation.lower() in candidate_titles:
                                        correct += 1
                                    else:
                                        logger.debug(
                                            f"Did not find `{relation.lower()}` in candidates"
                                        )
                                    total += 1
                        else:
                            sample["span_candidates"] = candidate_titles
                            # sample["window_candidates"] = candidate_titles
                            sample["span_candidates_scores"] = [
                                c.score for c in retrieved
                            ]
                            if log_recall:
                                candidate_titles_lower = [
                                    candidate.lower() for candidate in candidate_titles
                                ]
                                for ss, se, label in sample["window_labels"]:
                                    if label == "--NME--":
                                        continue
                                    if (
                                        label.replace("_", " ").lower()
                                        in candidate_titles_lower
                                    ):
                                        correct += 1
                                    else:
                                        logger.debug(
                                            f"Did not find `{label.replace('_', ' ').lower()}` in candidates"
                                        )
                                    total += 1
                        output_data.append(sample)

                    for sample in output_data:
                        f_out.write(json.dumps(sample) + "\n")

                    num_completed_docs += len(retrieved_accumulator)
                    retrieved_accumulator = []

            if len(retrieved_accumulator) > 0:
                output_data = []
                # get the correct document from the original dataset
                # the dataloader is not shuffled, so we can just count the number of
                # documents we have seen so far
                for sample, retrieved in zip(
                    samples[
                        num_completed_docs : num_completed_docs
                        + len(retrieved_accumulator)
                    ],
                    retrieved_accumulator,
                ):
                    candidate_titles = [
                        c.document.text
                        for c in retrieved  # TODO: add metadata if needed
                    ]
                    # TODO: compatibility shit
                    if relations:
                        sample["triplet_candidates"] = candidate_titles
                        sample["triplet_candidates_scores"] = [
                            c.score for c in retrieved
                        ]
                        if log_recall:
                            for triplet in sample["window_triplet_labels"]:
                                relation = triplet["relation"]
                                if relation.lower() in candidate_titles:
                                    correct += 1
                                else:
                                    logger.debug(
                                        f"Did not find `{relation.lower()}` in candidates"
                                    )
                                total += 1
                    else:
                        sample["span_candidates"] = candidate_titles
                        # sample["window_candidates"] = candidate_titles
                        sample["span_candidates_scores"] = [c.score for c in retrieved]
                        if log_recall:
                            for ss, se, label in sample["window_labels"]:
                                if label == "--NME--":
                                    continue
                                if label.replace("_", " ").lower() in candidate_titles:
                                    correct += 1
                                else:
                                    logger.debug(
                                        f"Did not find `{label.replace('_', ' ').lower()}` in candidates"
                                    )
                                total += 1
                    output_data.append(sample)

                for sample in output_data:
                    f_out.write(json.dumps(sample) + "\n")

                num_completed_docs += len(retrieved_accumulator)
                retrieved_accumulator = []

            end = time.time()
            logger.info(f"Retrieval took {end - start} seconds")
            if log_recall:
                recall = correct / total
                logger.info(f"Recall@{top_k}: {recall}")



import csv
import json
from pathlib import Path
import pickle
import sys
from typing import Any, Dict, List, Union

from relik.common.log import get_logger
from relik.common.utils import JsonSerializable

csv.field_size_limit(sys.maxsize)

logger = get_logger(__name__)


class Document:
    def __init__(
        self,
        text: str,
        id: int | None = None,
        metadata: Dict | None = None,
        **kwargs,
    ):
        self.text = text
        # if id is not provided, we use the hash of the text
        self.id = id if id is not None else hash(text)
        # if metadata is not provided, we use an empty dictionary
        self.metadata = metadata or {}

    def __str__(self):
        return f"{self.id}:{self.text}"

    def __repr__(self):
        return json.dumps(self.to_dict())

    def __eq__(self, other):
        if isinstance(other, Document):
            return self.id == other.id
        elif isinstance(other, int):
            return self.id == other
        elif isinstance(other, str):
            return self.text == other
        else:
            raise ValueError(
                f"Document must be compared with a Document, an int or a str, got `{type(other)}`"
            )

    def to_dict(self):
        return {"text": self.text, "id": self.id, "metadata": self.metadata}

    def to_json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, d: Dict):
        return cls(**d)

    @classmethod
    def from_file(cls, file_path: Union[str, Path], **kwargs):
        with open(file_path, "r") as f:
            d = json.load(f)
        return cls.from_dict(d)

    def save(self, file_path: Union[str, Path], **kwargs):
        with open(file_path, "w") as f:
            json.dump(self.to_dict(), f, indent=2)


class DocumentStore:
    """
    A document store is a collection of documents.

    Args:
        documents (:obj:`List[Document]`):
            The documents to store.
    """

    def __init__(self, documents: List[Document] = None) -> None:
        if documents is None:
            documents = []
        # if self.ingore_case:
        #     documents = [doc.lower() for doc in documents]
        self._documents = documents
        # build an index for the documents
        self._documents_index = {doc.id: doc for doc in self._documents}
        # build a reverse index for the documents
        self._documents_reverse_index = {doc.text: doc for doc in self._documents}

    def __len__(self):
        return len(self._documents)

    def __getitem__(self, index):
        return self._documents[index]

    def __iter__(self):
        return iter(self._documents)

    def __contains__(self, item):
        if isinstance(item, int):
            return item in self._documents_index
        elif isinstance(item, str):
            return item in self._documents_reverse_index
        elif isinstance(item, Document):
            return item.id in self._documents_index
        # return item in self._documents_index

    def __str__(self):
        return f"DocumentStore with {len(self)} documents"

    def __repr__(self):
        return self.__str__()

    def get_document_from_id(self, id: int) -> Document | None:
        """
        Retrieve a document by its ID.

        Args:
            id (`int`):
                The ID of the document to retrieve.

        Returns:
            Optional[Document]: The document with the given ID, or None if it does not exist.
        """
        if id not in self._documents_index:
            logger.warning(f"Document with id `{id}` does not exist, skipping")
        return self._documents_index.get(id, None)

    def get_document_from_text(self, text: str) -> Document | None:
        """
        Retrieve the document by its text.

        Args:
            text (`str`):
                The text of the document to retrieve.

        Returns:
            Optional[Document]: The document with the given text, or None if it does not exist.
        """
        if text not in self._documents_reverse_index:
            logger.warning(f"Document with text `{text}` does not exist, skipping")
        return self._documents_reverse_index.get(text, None)

    def get_document_from_index(self, index: int) -> Document | None:
        """
        Retrieve the document by its index.

        Args:
            index (`int`):
                The index of the document to retrieve.

        Returns:
            Optional[Document]: The document with the given index, or None if it does not exist.
        """
        if index >= len(self._documents):
            logger.warning(f"Document with index `{index}` does not exist, skipping")
        return self._documents[index]

    def add_documents(
        self, documents: List[Document] | List[str] | List[Dict]
    ) -> List[Document]:
        """
        Add a list of documents to the document store.

        Args:
            documents (`List[Document]`):
                The documents to add.

        Returns:
            List[Document]: The documents just added.
        """
        return [
            (
                self.add_document(Document.from_dict(doc))
                if isinstance(doc, Dict)
                else self.add_document(doc)
            )
            for doc in documents
        ]

    def add_document(
        self,
        text: str | Document,
        id: int | None = None,
        metadata: Dict | None = None,
    ) -> Document:
        """
        Add a document to the document store.

        Args:
            text (`str`):
                The text of the document to add.
            id (`int`, optional, defaults to None):
                The ID of the document to add.
            metadata (`Dict`, optional, defaults to None):
                The metadata of the document to add.

        Returns:
            Document: The document just added.
        """
        if isinstance(text, str):
            # check if the document already exists
            if text in self:
                logger.warning(f"Document `{text}` already exists, skipping")
                return self._documents_reverse_index[text]
            if id is None:
                # get the len of the documents and add 1
                id = len(self._documents)  # + 1
            text = Document(text, id, metadata)

        if text in self:
            logger.warning(f"Document `{text}` already exists, skipping")
            return self._documents_index[text.id]

        self._documents.append(text)
        self._documents_index[text.id] = text
        self._documents_reverse_index[text.text] = text
        return text
        # if id in self._documents_index:
        #     logger.warning(f"Document with id `{id}` already exists, skipping")
        #     return self._documents_index[id]
        # if text_or_document in self._documents_reverse_index:
        #     logger.warning(f"Document with text `{text_or_document}` already exists, skipping")
        #     return self._documents_reverse_index[text_or_document]
        # self._documents.append(Document(text_or_document, id, metadata))
        # self._documents_index[id] = self._documents[-1]
        # self._documents_reverse_index[text_or_document] = self._documents[-1]
        # return self._documents_index[id]

    def delete_document(self, document: int | str | Document) -> bool:
        """
        Delete a document from the document store.

        Args:
            document (`int`, `str` or `Document`):
                The document to delete.

        Returns:
            bool: True if the document has been deleted, False otherwise.
        """
        if isinstance(document, int):
            return self.delete_by_id(document)
        elif isinstance(document, str):
            return self.delete_by_text(document)
        elif isinstance(document, Document):
            return self.delete_by_document(document)
        else:
            raise ValueError(
                f"Document must be an int, a str or a Document, got `{type(document)}`"
            )

    def delete_by_id(self, id: int) -> bool:
        """
        Delete a document by its ID.

        Args:
            id (`int`):
                The ID of the document to delete.

        Returns:
            bool: True if the document has been deleted, False otherwise.
        """
        if id not in self._documents_index:
            logger.warning(f"Document with id `{id}` does not exist, skipping")
            return False
        del self._documents_reverse_index[self._documents_index[id]]
        del self._documents_index[id]
        return True

    def delete_by_text(self, text: str) -> bool:
        """
        Delete a document by its text.

        Args:
            text (`str`):
                The text of the document to delete.

        Returns:
            bool: True if the document has been deleted, False otherwise.
        """
        if text not in self._documents_reverse_index:
            logger.warning(f"Document with text `{text}` does not exist, skipping")
            return False
        del self._documents_reverse_index[text]
        del self._documents_index[self._documents_index[text]]
        return True

    def delete_by_document(self, document: Document) -> bool:
        """
        Delete a document by its text.

        Args:
            document (:obj:`Document`):
                The document to delete.

        Returns:
            bool: True if the document has been deleted, False otherwise.
        """
        if document.id not in self._documents_index:
            logger.warning(f"Document {document} does not exist, skipping")
            return False
        del self._documents[self._documents.index(document)]
        del self._documents_index[document.id]
        del self._documents_reverse_index[self._documents_index[document.id]]

    def to_dict(self):
        return [doc.to_dict() for doc in self._documents]

    @classmethod
    def from_dict(cls, d):
        return cls([Document.from_dict(doc) for doc in d])

    @classmethod
    def from_file(
        cls, file_path: Union[str, Path], skip_metadata: bool = False, **kwargs
    ):
        """
        Load documents from a file.

        Args:
            file_path (Union[str, Path]): The path to the file containing the documents.
            skip_metadata (bool, optional): Whether to skip loading metadata for each document. Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            cls: An instance of the class with the loaded documents.
        """
        with open(file_path, "r") as f:
            docs = []
            for line in f:
                doc = json.loads(line)
                if skip_metadata:
                    doc.pop("metadata", None)
                docs.append(Document.from_dict(doc))
        return cls(docs)

    @classmethod
    def from_pickle(cls, file_path: Union[str, Path], **kwargs):
        with open(file_path, "rb") as handle:
            d = pickle.load(handle)
        return cls(d)

    @classmethod
    def from_tsv(
        cls,
        file_path: Union[str, Path],
        ingore_case: bool = False,
        delimiter: str = "\t",
        **kwargs,
    ):
        d = []
        # load a tsv/csv file and take the header into account
        # the header must be `id\ttext\t[list of metadata keys]`
        with open(file_path, "r", encoding="utf8") as f:
            csv_reader = csv.reader(f, delimiter=delimiter, **kwargs)
            header = next(csv_reader)
            id, text, *metadata_keys = header
            for i, row in enumerate(csv_reader):
                # check if id can be casted to int
                # if not, we add it to the metadata and use `i` as id
                try:
                    s_id = int(row[header.index(id)])
                    row_metadata_keys = metadata_keys
                except ValueError:
                    row_metadata_keys = [id] + metadata_keys
                    s_id = i

                d.append(
                    Document(
                        text=(
                            row[header.index(text)].strip().lower()
                            if ingore_case
                            else row[header.index(text)].strip()
                        ),
                        id=s_id,  # row[header.index(id)],
                        metadata={
                            key: row[header.index(key)] for key in row_metadata_keys
                        },
                    )
                )
        return cls(d)

    def save(self, file_path: Union[str, Path], **kwargs):
        with open(file_path, "w") as f:
            for doc in self._documents:
                # save as json lines
                f.write(json.dumps(doc.to_dict()) + "\n")



import argparse
from pprint import pprint
from typing import Optional

from relik.reader.pytorch_modules.span import RelikReaderForSpanExtraction
from relik.reader.utils.strong_matching_eval import StrongMatching
from relik.reader.data.relik_reader_sample import load_relik_reader_samples


def predict(
    model_path: str,
    dataset_path: str,
    token_batch_size: int,
    is_eval: bool,
    output_path: Optional[str],
) -> None:
    relik_reader = RelikReaderForSpanExtraction(
        model_path, dataset_kwargs={"use_nme": True}, device="cuda"
    )
    samples = list(load_relik_reader_samples(dataset_path))
    predicted_samples = relik_reader.read(
        samples=samples, token_batch_size=token_batch_size, progress_bar=True
    )
    if is_eval:
        eval_dict = StrongMatching()(predicted_samples)
        pprint(eval_dict)
    if output_path is not None:
        with open(output_path, "w") as f:
            for sample in predicted_samples:
                f.write(sample.to_jsons() + "\n")


def parse_arg() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-path",
        required=True,
    )
    parser.add_argument("--dataset-path", "-i", required=True)
    parser.add_argument("--is-eval", action="store_true")
    parser.add_argument(
        "--output-path",
        "-o",
    )
    parser.add_argument("--token-batch-size", default=4096)
    return parser.parse_args()


def main():
    args = parse_arg()
    predict(
        args.model_path,
        args.dataset_path,
        token_batch_size=args.token_batch_size,
        is_eval=args.is_eval,
        output_path=args.output_path,
    )


if __name__ == "__main__":
    main()
